import logging
import re
from copy import (
    deepcopy,
)

from pydantic import (
    BaseModel,
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.java.model import (
    JavaArchive,
    JavaPomProject,
)
from labels.pkg.cataloger.java.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)
# Constants
QUOTE = r'["\']'
NL = r"(\n?\s*)?"
TEXT = r"[a-zA-Z0-9._-]+"
RE_GRADLE_KTS: re.Pattern[str] = re.compile(
    rf"(runtimeOnly|api|compile|plugin|compileOnly|implementation)\({QUOTE}"
    rf"(?P<group>{TEXT}):(?P<name>{TEXT}):"
    rf"(?P<version>{TEXT})"
    rf"(?::(?P<classifier>{TEXT}))?"
    rf"{QUOTE}\)",
)


def is_comment(line: str) -> bool:
    return (
        line.strip().startswith("//")
        or line.strip().startswith("/*")
        or line.strip().endswith("*/")
    )


class LockFileDependency(BaseModel):
    group: str
    name: str
    version: str
    line: int | None = None


def parse_gradle_lockfile_kts(
    _resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    dependencies = parse_dependencies(reader)
    packages = create_packages(dependencies, reader.location)
    return packages, []


def parse_dependencies(reader: LocationReadCloser) -> list[LockFileDependency]:
    dependencies: list[LockFileDependency] = []
    is_block_comment = False

    for line_no, raw_line in enumerate(reader.read_closer.readlines(), start=1):
        line = raw_line.strip()
        is_block_comment = update_block_comment_status(line, is_block_comment=is_block_comment)

        if is_block_comment or is_comment(line):
            continue

        dependency = extract_dependency(line, line_no)
        if dependency:
            dependencies.append(dependency)

    return dependencies


def update_block_comment_status(line: str, *, is_block_comment: bool) -> bool:
    if "/*" in line:
        return True
    if "*/" in line:
        return False
    return is_block_comment


def extract_dependency(line: str, line_no: int) -> LockFileDependency | None:
    if match := RE_GRADLE_KTS.match(line):
        version = match.group("version")
        if version:
            return LockFileDependency(
                group=match.group("group"),
                name=match.group("name"),
                version=version,
                line=line_no,
            )

    return None


def create_packages(
    dependencies: list[LockFileDependency],
    reader_location: Location,
) -> list[Package]:
    packages: list[Package] = []

    for dependency in dependencies:
        name = dependency.name
        version = dependency.version

        if not name or not version:
            continue

        location = deepcopy(reader_location)
        if location.coordinates:
            location.coordinates.line = dependency.line

        archive = create_java_archive(dependency, name, version)

        try:
            package = create_package(name, version, location, archive, dependency.group)
            packages.append(package)
        except ValidationError as ex:
            log_validation_error(ex, location)

    return packages


def create_java_archive(
    dependency: LockFileDependency,
    name: str,
    version: str,
) -> JavaArchive:
    return JavaArchive(
        pom_project=JavaPomProject(
            group_id=dependency.group,
            name=name,
            artifact_id=name,
            version=version,
        ),
    )


def create_package(
    name: str,
    version: str,
    location: Location,
    archive: JavaArchive,
    group: str,
) -> Package:
    return Package(
        name=f"{group}:{name}",
        version=version,
        locations=[location],
        language=Language.JAVA,
        type=PackageType.JavaPkg,
        metadata=archive,
        p_url=package_url(name, version, archive),
        licenses=[],
    )


def log_validation_error(ex: ValidationError, location: Location) -> None:
    LOGGER.warning(
        "Malformed package. Required fields are missing or data types are incorrect.",
        extra={
            "extra": {
                "exception": format_exception(str(ex)),
                "location": location.path(),
            },
        },
    )



import glob
import logging
import os
import re
from collections.abc import Generator
from fnmatch import (
    translate,
)
from pathlib import Path
from typing import (
    TextIO,
    cast,
)

import magic
from pydantic import (
    BaseModel,
)

from labels.file.coordinates import (
    Coordinates,
)
from labels.file.location import (
    Location,
)
from labels.file.metadata import (
    Metadata,
)
from labels.file.resolver import (
    FileReader,
    Resolver,
)
from labels.file.type import (
    Type,
)

LOGGER = logging.getLogger(__name__)


def _normalize_rules(rules: tuple[str, ...]) -> tuple[str, ...]:
    normalized_rules: list[str] = []
    for rule in rules:
        new_rule = rule
        if new_rule.startswith("glob(") and new_rule.endswith(")"):
            new_rule = new_rule[5:-1]
        if new_rule == ".":
            new_rule = "**"
        if new_rule.endswith("/"):
            new_rule += "**"
        if new_rule.startswith("**/"):
            normalized_rules.append(new_rule[3:])
        normalized_rules.append(new_rule)
    return tuple(normalized_rules)


class Directory(Resolver, BaseModel):
    root: str
    exclude: tuple[str, ...]

    def __post__init__(self, root: str, exclude: tuple[str, ...]) -> None:
        self.root = os.path.realpath(Path(root).resolve())
        self.exclude = exclude

    def has_path(self, path: str) -> bool:
        path = os.path.join(self.root, path.lstrip("/"))
        return Path(path).exists()

    def files_by_path(self, *paths: str) -> list[Location]:
        locations: list[Location] = []
        for path in paths:
            relative_path = path.replace(self.root, "").lstrip("/")
            full_path = os.path.join(self.root, relative_path)
            if Path(full_path).exists():
                locations.append(
                    Location(
                        coordinates=Coordinates(real_path=full_path, file_system_id=""),
                        access_path=relative_path,
                        annotations={},
                    ),
                )
        return locations

    def files_by_glob(self, *patters: str) -> list[Location]:
        result: list[Location] = []
        for pattern in patters:
            result.extend(
                Location(
                    coordinates=Coordinates(
                        real_path=os.path.join(self.root, item),
                        file_system_id="",
                    ),
                    access_path=item,
                    annotations={},
                )
                for item in glob.glob(pattern, root_dir=self.root, recursive=True)  # noqa: PTH207
            )

        return result

    def files_by_mime_type(self, mime_type: str) -> list[Location]:
        matching_files = []
        mime_detector = magic.Magic(mime=True)

        for dirpath, _, filenames in os.walk(self.root):
            for filename in filenames:
                relative_path = os.path.join(dirpath, filename).replace(self.root, "").lstrip("/")
                result_mime_type = mime_detector.from_file(relative_path)
                if mime_type == result_mime_type:
                    matching_files.append(
                        Location(
                            coordinates=Coordinates(
                                real_path=os.path.join(self.root, relative_path),
                                file_system_id="",
                            ),
                            access_path=relative_path,
                            annotations={},
                        ),
                    )

        return matching_files

    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: FileReader | None = None,
        mode: str | None = None,
    ) -> TextIO | None:
        if (
            location.coordinates
            and location.coordinates.real_path is not None
            and Path(location.coordinates.real_path).exists()
        ):
            return cast(
                TextIO,
                (function_reader or open)(
                    location.coordinates.real_path,
                    encoding="utf-8",
                    mode=mode or "r",
                ),
            )

        return None

    def file_metadata_by_location(self, location: Location) -> Metadata | None:
        link_destination = None
        if not location.access_path:
            return None

        stats = Path(location.access_path).stat()

        if Path(location.access_path).is_symlink():
            file_type = Type.TYPE_SYM_LINK
            link_destination = str(Path(location.access_path).readlink())
        elif Path(location.access_path).is_dir():
            file_type = Type.TYPE_DIRECTORY
        elif Path(location.access_path).is_file():
            file_type = Type.TYPE_REGULAR
        else:
            file_type = Type.TYPE_IRREGULAR

        mime_type = magic.Magic(mime=True).from_file(location.access_path)

        return Metadata(
            path=location.access_path,
            link_destination=link_destination or "",
            user_id=stats.st_uid,
            group_id=stats.st_gid,
            type=file_type,
            mime_type=mime_type,
        )

    def relative_file_path(self, _: Location, _path: str) -> Location:
        return Location(
            coordinates=Coordinates(real_path=_path, file_system_id="", line=None),
            access_path=_path.replace(self.root, "").lstrip("/"),
        )

    def walk_file(self) -> Generator[str, None, None]:
        excluded_dirs = ["node_modules", "dist", "__pycache__"]
        exclude_regex = [
            translate(os.path.join(self.root, rule)) for rule in _normalize_rules(self.exclude)
        ]

        for dirpath, _, filenames in os.walk(self.root):
            if any(
                (
                    dirpath.endswith(excluded_dir)
                    or f"{os.path.sep}{excluded_dir}{os.path.sep}" in dirpath
                )
                for excluded_dir in excluded_dirs
            ):
                continue

            for filename in filenames:
                full_path = os.path.join(dirpath, filename)
                if any(re.match(regex, full_path) for regex in exclude_regex):
                    continue

                relative_path = full_path.replace(self.root, "").lstrip("/")
                yield relative_path


class DirectoryConfig(BaseModel):
    path: str
    exclude: tuple[str, ...]


class DirectorySource(BaseModel):
    config: DirectoryConfig
    resolver: Directory | None = None

    def file_resolver(self) -> Resolver:
        if self.resolver is None:
            self.resolver = Directory(root=self.config.path, exclude=self.config.exclude)
        return self.resolver


def new_from_directory_path(path: str, exclude: tuple[str, ...]) -> DirectorySource | None:
    return new_from_directory(DirectoryConfig(path=path, exclude=exclude))


def new_from_directory(cfg: DirectoryConfig) -> DirectorySource | None:
    if not Path(cfg.path).is_dir():
        LOGGER.error("Given path is not a directory: %s", cfg.path)
        return None

    return DirectorySource(config=cfg, resolver=None)



import os
import re

import bugsnag
from bugsnag.notification import (
    Notification,
)

from labels.context import (
    BASE_DIR,
    CI_COMMIT_SHORT_SHA,
)
from labels.utils import (
    env,
)


def _remove_nix_hash(path: str) -> str:
    pattern = r"(\/nix\/store\/[a-z0-9]{32}-)"
    result = re.search(pattern, path)
    if not result:
        return path
    return path[result.end(0) :]


def bugsnag_remove_nix_hash(
    notification: Notification,
) -> None:
    notification.stacktrace = [
        {**trace, "file": _remove_nix_hash(trace["file"])} for trace in notification.stacktrace
    ]


def bugsnag_add_batch_metadata(
    notification: Notification,
) -> None:
    if batch_job_id := os.environ.get("AWS_BATCH_JOB_ID"):
        batch_job_info = {"batch_job_id": batch_job_id}
        notification.add_tab("batch_job_info", batch_job_info)


def initialize_bugsnag() -> None:
    bugsnag.before_notify(bugsnag_add_batch_metadata)
    bugsnag.before_notify(bugsnag_remove_nix_hash)
    bugsnag.configure(
        notify_release_stages=["production"],
        release_stage=env.guess_environment(),
        app_version=CI_COMMIT_SHORT_SHA,
        project_root=BASE_DIR,
        send_environment=True,
    )
    bugsnag.start_session()



import logging
from datetime import datetime

from defusedxml import ElementTree
from pydantic import BaseModel, ConfigDict

from labels.internal.package_information.api_interface import make_get

LOGGER = logging.getLogger(__name__)


class MavenSearchDocResponse(BaseModel):
    id_: str
    group: str
    artifact: str
    version: str
    timestamp: int
    extra_classifiers: list[str]
    packaging: str | None = None
    tags: list[str] | None = None
    model_config = ConfigDict(frozen=True)


class AuthorsLicenses(BaseModel):
    authors: list[str]
    licenses: list[str]
    model_config = ConfigDict(frozen=True)


class MavenPackageInfo(BaseModel):
    group_id: str
    artifact_id: str
    latest_version: str | None = None
    release_date: int | None = None
    authors: list[str] | None = None
    licenses: list[str] | None = None
    version: str | None = None
    jar_url: str | None = None
    hash: str | None = None
    model_config = ConfigDict(frozen=True)


def search_maven_package(artifact_id: str, version: str) -> MavenSearchDocResponse | None:
    base_url = "https://search.maven.org/solrsearch/select"

    query = f"a:{artifact_id} AND v:{version}" if version else f"a:{artifact_id}"

    params: dict[str, str | int] = {"q": query, "rows": 5, "wt": "json"}

    package_data = make_get(base_url, params=params, timeout=30)
    if package_data:
        docs = package_data["response"].get("docs", [])

        if not docs or len(docs) > 1:
            return None

        try:
            result = MavenSearchDocResponse(
                id_=docs[0]["id"],
                group=docs[0]["g"],
                artifact=docs[0]["a"],
                version=docs[0]["v"] if "v" in docs[0] else docs[0]["latestVersion"],
                packaging=docs[0].get("p", None),
                timestamp=int(docs[0]["timestamp"] // 1000),
                extra_classifiers=docs[0]["ec"],
                tags=docs[0].get("tags", None),
            )
        except KeyError as exc:
            LOGGER.exception(
                "Error parsing Maven search response",
                extra={
                    "extra": {
                        "artifact_id": artifact_id,
                        "version": version,
                        "doc": docs[0],
                        "key": exc.args[0],
                    },
                },
            )
            return None
        return result

    return None


def _get_latest_version_authors_licenses(
    group_id: str,
    artifact_id: str,
    latest_version: str,
) -> AuthorsLicenses | None:
    group_id_path = group_id.replace(".", "/")
    # Fetch the latest version's POM file
    pom_url = (
        f"https://repo1.maven.org/maven2"
        f"/{group_id_path}/{artifact_id}/{latest_version}"
        f"/{artifact_id}-{latest_version}.pom"
    )
    response = make_get(pom_url, content=True, timeout=30)
    if response:
        pom_xml = ElementTree.fromstring(response)
        namespace = {"m": "http://maven.apache.org/POM/4.0.0"}

        # Extract authors and licenses
        authors = [
            author.text
            for author in pom_xml.findall("m:developers/m:developer/m:name", namespace)
            if author.text
        ]
        licenses = [
            getattr(package_license.find("m:name", namespace), "text", "")
            for package_license in pom_xml.findall("m:licenses/m:license", namespace)
        ]
        return AuthorsLicenses(authors=authors, licenses=licenses)
    return None


def _get_all_package(group_id: str, artifact_id: str) -> MavenPackageInfo | None:
    group_id_path = group_id.replace(".", "/")
    # URL for the metadata file
    metadata_url = (
        f"https://repo1.maven.org/maven2/{group_id_path}/{artifact_id}/maven-metadata.xml"
    )
    response = make_get(metadata_url, content=True, timeout=30)
    if response:
        # Parse the metadata XML response
        metadata_xml = ElementTree.fromstring(response)
        latest_version = metadata_xml.find("versioning/latest")
        release_date_tag = metadata_xml.find("versioning/lastUpdated")
        release_date: int | None = None
        if release_date_tag is not None and release_date_tag.text:
            release_date = int(
                datetime.strptime(release_date_tag.text + "+0000", "%Y%m%d%H%M%S%z").timestamp(),
            )

        authors_licenses = (
            _get_latest_version_authors_licenses(group_id, artifact_id, latest_version.text or "")
            if latest_version is not None
            else None
        )
        return MavenPackageInfo(
            group_id=group_id,
            artifact_id=artifact_id,
            latest_version=latest_version.text if latest_version is not None else None,
            release_date=release_date,
            authors=authors_licenses.authors if authors_licenses else None,
            licenses=authors_licenses.licenses if authors_licenses else None,
        )

    return None


def get_maven_package_info(
    group_id: str,
    artifact_id: str,
    version: str | None = None,
) -> MavenPackageInfo | None:
    group_id_path = group_id.replace(".", "/")

    if version:
        # URL for the specific version's POM file
        pom_url = (
            f"https://repo1.maven.org/maven2/{group_id_path}"
            f"/{artifact_id}/{version}/{artifact_id}-{version}.pom"
        )
        response = make_get(pom_url, timeout=30, content=True)
        if response:
            # Extract the download URL and hash
            jar_url = (
                f"https://repo1.maven.org/maven2/{group_id_path}"
                f"/{artifact_id}/{version}/{artifact_id}-{version}.jar"
            )
            hash_url = (
                f"https://repo1.maven.org/maven2/{group_id_path}"
                f"/{artifact_id}/{version}/{artifact_id}-{version}.jar.sha1"
            )
            hash_response: str | None = make_get(hash_url, content=True, timeout=30)

            package_hash = hash_response.strip() if hash_response else "Hash not available"

            return MavenPackageInfo(
                group_id=group_id,
                artifact_id=artifact_id,
                version=version,
                jar_url=jar_url,
                hash=package_hash,
            )

        return None

    return _get_all_package(group_id, artifact_id)



from fnmatch import (
    fnmatch,
)

import reactivex
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)

from labels.pkg.cataloger.generic.cataloger import (
    Request,
)
from labels.pkg.cataloger.python.parse_pipfile_deps import (
    parse_pipfile_deps,
)
from labels.pkg.cataloger.python.parse_pipfile_lock import (
    parse_pipfile_lock_deps,
)
from labels.pkg.cataloger.python.parse_poetry_lock import (
    parse_poetry_lock,
)
from labels.pkg.cataloger.python.parse_pyproject_toml import (
    parse_pyproject_toml,
)
from labels.pkg.cataloger.python.parse_requirements import (
    parse_requirements_txt,
)
from labels.pkg.cataloger.python.parse_wheel_egg import (
    parse_wheel_or_egg,
)


def on_next_python(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if fnmatch(value, "*requirements*.txt"):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_requirements_txt,
                            parser_name="python-requirements-cataloger",
                        ),
                    )
                elif any(
                    fnmatch(value, x) for x in ("*poetry.lock", "poetry.lock", "*/poetry.lock")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_poetry_lock,
                            parser_name="python-poetry-lock-cataloger",
                        ),
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        "**/*.egg-info",
                        "**/*dist-info/METADATA",
                        "**/*egg-info/PKG-INFO",
                        "**/*DIST-INFO/METADATA",
                        "**/*EGG-INFO/PKG-INFO",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_wheel_or_egg,
                            parser_name="python-installed-package-cataloger",
                        ),
                    )
                elif any(
                    fnmatch(value, pattern) for pattern in ("**/Pipfile.lock", "Pipfile.lock")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_pipfile_lock_deps,
                            parser_name="python-pipfile-lock-cataloger",
                        ),
                    )
                elif any(fnmatch(value, pattern) for pattern in ("**/Pipfile", "Pipfile")):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_pipfile_deps,
                            parser_name="python-pipfile-package-cataloger",
                        ),
                    )
                elif any(
                    fnmatch(value, pattern) for pattern in ("**/pyproject.toml", "pyproject.toml")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_pyproject_toml,
                            parser_name="python-pyproject-toml-cataloger",
                        ),
                    )
            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



import logging
import multiprocessing
import traceback
from collections.abc import Callable
from typing import (
    cast,
)

import reactivex
from reactivex import (
    Observable,
)
from reactivex import operators as ops
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)
from reactivex.scheduler import (
    ThreadPoolScheduler,
)

from labels.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from labels.file.resolver import (
    Resolver,
)
from labels.linux.release import (
    identify_release,
)
from labels.model.core import (
    Package,
)
from labels.pkg.cataloger.generic.cataloger import (
    Request,
    Task,
    execute_parsers,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.handle import (
    handle_parser,
)
from labels.pkg.cataloger.python.model import (
    PythonPackage,
)

LOGGER = logging.getLogger(__name__)


def strip_version_specifier(item: str) -> str:
    # Define the characters that indicate the start of a version specifier
    specifiers = "[(<>="

    # Find the index of the first occurrence of any specifier character
    index = next((i for i, char in enumerate(item) if char in specifiers), None)

    # If no specifier character is found, return the original string
    if index is None:
        return item.strip()

    # Return the substring up to the first specifier character, stripped of
    # leading/trailing whitespace
    return item[:index].strip()


def handle_relationships(packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []
    for package in packages:
        match package.found_by:
            case "python-installed-package-cataloger":
                python_package: PythonPackage = cast(PythonPackage, package.metadata)
                for dep in python_package.dependencies if python_package.dependencies else []:
                    dep_name = strip_version_specifier(dep)
                    if dep_package := next((x for x in packages if x.name == dep_name), None):
                        relationships.append(
                            Relationship(
                                from_=dep_package,
                                to_=package,
                                type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                            ),
                        )
    return relationships


def gen_location(
    resolver: Resolver,
) -> Callable[[Observable[Request]], Observable]:
    def _handle(source: Observable[Request]) -> Observable:
        def subscribe(
            observer: ObserverBase[Task],
            scheduler: SchedulerBase | None = None,
        ) -> reactivex.abc.DisposableBase:
            def on_next(value: Request) -> None:
                try:
                    locations = resolver.files_by_path(value.real_path)
                    for location in locations:
                        observer.on_next(
                            Task(
                                location=location,
                                parser=value.parser,
                                parser_name=value.parser_name,
                            ),
                        )
                except (
                    Exception  # noqa: BLE001
                ) as ex:
                    observer.on_error(ex)

            return source.subscribe(
                on_next,
                observer.on_error,
                observer.on_completed,
                scheduler=scheduler,
            )

        return reactivex.create(subscribe)

    return _handle


def log_and_continue(e: Exception, file_item: str) -> Observable[None]:
    LOGGER.error(
        "Error found while resolving packages of %s: %s: %s",
        file_item,
        str(e),
        traceback.format_exc(),
    )
    return reactivex.empty()


def process_file_item(
    file_item: str,
    resolver: Resolver,
    pool_scheduler: ThreadPoolScheduler,
) -> Observable[tuple[list[Package], list[Relationship]]]:
    def factory(_: SchedulerBase | None = None) -> Observable[str]:
        return reactivex.just(file_item)

    return reactivex.defer(factory).pipe(
        handle_parser(scheduler=pool_scheduler),
        gen_location(resolver),
        execute_parsers(resolver, Environment(linux_release=identify_release(resolver))),
        ops.catch(lambda e, _: log_and_continue(e, file_item)),
    )


def package_operations_factory(
    resolver: Resolver,
) -> tuple[list[Package], list[Relationship]]:
    observer: Observable[str] = reactivex.from_iterable(resolver.walk_file())
    result_packages: list[Package] = []
    result_relations: list[Relationship] = []
    completed_event = multiprocessing.Event()
    errors = []

    def on_completed() -> None:
        completed_event.set()

    def on_error(error: Exception) -> None:
        errors.append(error)
        on_completed()

    def on_next(value: tuple[list[Package], list[Relationship]]) -> None:
        packages, relations = value
        result_packages.extend(packages)
        result_relations.extend(relations)

    optimal_thread_count = multiprocessing.cpu_count()
    pool_scheduler = ThreadPoolScheduler(optimal_thread_count)

    final_obs: Observable[tuple[list[Package], list[Relationship]]] = observer.pipe(
        ops.map(
            lambda file_item: process_file_item(file_item, resolver, pool_scheduler),  # type: ignore[arg-type]
        ),
        ops.merge(max_concurrent=optimal_thread_count),
    )
    final_obs.subscribe(on_next=on_next, on_error=on_error, on_completed=on_completed)

    completed_event.wait()
    result_relations.extend(handle_relationships(result_packages))

    return result_packages, result_relations



import re
from typing import TextIO, cast


def parse_licenses_from_copyright(reader: TextIO) -> list[str]:
    result = []
    for raw_line in reader:
        line = raw_line.rstrip("\n")
        if value := find_license_clause(r"^License: (?P<license>\S*)", "license", line):
            result.append(value)
        if value := find_license_clause(
            r"/usr/share/common-licenses/(?P<license>[0-9A-Za-z_.\-]+)",
            "license",
            line,
        ):
            result.append(value)

    return list(set(result))


def find_license_clause(pattern: str, value_group: str, line: str) -> str | None:
    match = re.search(pattern, line)
    if match:
        return ensure_is_single_license(cast(str, match.group(value_group)))
    return None


def ensure_is_single_license(candidate: str) -> str | None:
    candidate = candidate.strip()
    if " or " in candidate or " and " in candidate:
        # This is a multi-license summary, ignore this.
        return None
    if candidate and candidate.lower() != "none":
        # The license may be at the end of a sentence; clean '.' characters.
        return candidate.rstrip(".")
    return None



import logging
from collections.abc import (
    Iterator,
)
from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from more_itertools import (
    flatten,
)
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from labels.internal.collection.types import (
    IndexedDict,
    ParsedValue,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)
EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def _get_relationships(
    packages: list[Package],
    dependencies: dict[str, ParsedValue],
) -> list[Relationship]:
    results: list[Relationship] = []
    for package_name, depens in dependencies.items():
        if not isinstance(depens, IndexedDict):
            continue
        deps = (p for p in packages for dep_name in depens if dep_name == p.name)
        if current_package := next(
            (x for x in packages if x.name == package_name),
            None,
        ):
            results.extend(
                Relationship(
                    from_=x,
                    to_=current_package,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
                for x in deps
            )
    return results


def parse_dotnet_package_lock(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    packages: list[Package] = []
    target_deps: ParsedValue = package_json.get("dependencies", IndexedDict())
    if not isinstance(target_deps, IndexedDict):
        LOGGER.warning("No deps found in package JSON")
        return ([], [])
    dependencies: dict[str, ParsedValue] = {}

    for package_name, package_value in cast(
        Iterator[IndexedDict[str, ParsedValue]],
        flatten(x.items() for x in target_deps.values() if isinstance(x, IndexedDict)),
    ):
        if not isinstance(package_value, IndexedDict):
            continue

        is_transitive: str = package_value.get("type", "") == "Transitive"

        version: str | None = package_value.get("resolved")

        if not package_name or not version:
            continue

        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = package_value.position.start.line
            location.dependency_type = (
                DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
            )

        dependencies[package_name] = package_value.get("dependencies", EMPTY_DICT)
        try:
            packages.append(
                Package(
                    name=package_name,
                    version=version,
                    locations=[location],
                    licenses=[],
                    type=PackageType.DotnetPkg,
                    language=Language.DOTNET,
                    metadata=None,
                    p_url=PackageURL(
                        type="nuget",
                        namespace="",
                        name=package_name,
                        version=version,
                        qualifiers={},
                        subpath="",
                    ).to_string(),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    relationships: list[Relationship] = _get_relationships(packages, dependencies)

    return packages, relationships



import struct

from pydantic import (
    BaseModel,
)

from labels.pkg.cataloger.redhat.rpmdb.inet import (
    htonl,
    htonlu,
)
from labels.pkg.cataloger.redhat.rpmdb.rpmtags import (
    HEADER_I18NTABLE,
    RPM_BIN_TYPE,
    RPM_I18NSTRING_TYPE,
    RPM_MAX_TYPE,
    RPM_MIN_TYPE,
    RPM_STRING_ARRAY_TYPE,
    RPM_STRING_TYPE,
    RPMTAG_HEADERI18NTABLE,
    RPMTAG_HEADERIMAGE,
    RPMTAG_HEADERIMMUTABLE,
    RPMTAG_HEADERSIGNATURES,
)

REGION_TAG_COUNT = 16
REGION_TAG_TYPE = RPM_BIN_TYPE

HEADER_MAX_BYTES = 256 * 1024 * 1024

TYPE_SIZES = [
    0,  # RPM_NULL_TYPE
    1,  # RPM_CHAR_TYPE
    1,  # RPM_INT8_TYPE
    2,  # RPM_INT16_TYPE
    4,  # RPM_INT32_TYPE
    8,  # RPM_INT64_TYPE
    -1,  # RPM_STRING_TYPE
    1,  # RPM_BIN_TYPE
    -1,  # RPM_STRING_ARRAY_TYPE
    -1,  # RPM_I18NSTRING_TYPE
    0,
    0,
    0,
    0,
    0,
    0,
]

TYPE_ALIGN = [
    1,  # RPM_NULL_TYPE
    1,  # RPM_CHAR_TYPE
    1,  # RPM_INT8_TYPE
    2,  # RPM_INT16_TYPE
    4,  # RPM_INT32_TYPE
    8,  # RPM_INT64_TYPE
    1,  # RPM_STRING_TYPE
    1,  # RPM_BIN_TYPE
    1,  # RPM_STRING_ARRAY_TYPE
    1,  # RPM_I18NSTRING_TYPE
    0,
    0,
    0,
    0,
    0,
    0,
]


class EntryInfo(BaseModel):
    tag: int  # int32 in Go
    type: int  # uint32 in Go
    offset: int  # int32 in Go
    count: int  # uint32 in Go


class IndexEntry(BaseModel):
    info: EntryInfo
    length: int
    rd_len: int
    data: bytes


class HdrBlob(BaseModel):
    pe_list: list[EntryInfo]  # List of `EntryInfo`
    i_length: int  # int32 in Go
    d_length: int  # int32 in Go
    pv_len: int  # int32 in Go
    data_start: int  # int32 in Go
    data_end: int  # int32 in Go
    region_tag: int  # int32 in Go
    ril: int  # int32 in Go
    rdl: int  # int32 in Go


def hdrchk_range(d_length: int, offset: int) -> bool:
    return offset < 0 or offset > d_length


def hdrchk_tag(tag: int) -> bool:
    return tag < HEADER_I18NTABLE


def hdrchk_type(type_: int) -> bool:
    return type_ < RPM_MIN_TYPE or type_ > RPM_MAX_TYPE


def hdrchk_align(type_: int, offset: int) -> bool:
    return offset & (TYPE_ALIGN[type_] - 1) != 0


def ei2h(package_entry: EntryInfo) -> EntryInfo:
    return EntryInfo(
        type=htonlu(package_entry.type),  # Convert unsigned 32-bit integer to network byte order
        count=htonlu(package_entry.count),  # Convert unsigned 32-bit integer to network byte order
        offset=htonl(package_entry.offset),  # Convert signed 32-bit integer to network byte order
        tag=htonl(package_entry.tag),  # Convert signed 32-bit integer to network byte order
    )


def data_length(data: bytes, type_: int, count: int, data_start: int, data_end: int) -> int:
    length = 0

    if type_ == RPM_STRING_TYPE:
        if count != 1:
            return -1
        length = strtaglen(data, 1, data_start, data_end)
    elif type_ in [RPM_STRING_ARRAY_TYPE, RPM_I18NSTRING_TYPE]:
        length = strtaglen(data, count, data_start, data_end)
    else:
        if TYPE_SIZES[type_] == -1:
            return -1
        length = TYPE_SIZES[type_ & 0xF] * count
        if length < 0 or data_end > 0 < data_start + length - data_end:
            return -1

    return length


def strtaglen(data: bytes, count: int, dat_start: int, data_end: int) -> int:
    length = 0
    if dat_start >= data_end or data_end > len(data):
        return -1

    for _ in range(count):
        offset = dat_start + length
        if offset > len(data):
            return -1
        try:
            index = data[offset:data_end].index(0x00)
        except ValueError:
            # Null terminator not found within the specified bounds
            return -1
        length += index + 1  # Include the null terminator

    return length


def align_diff(type_: int, align_size: int) -> int:
    type_size = TYPE_SIZES[type_]
    if type_size > 1:
        diff = type_size - (align_size % type_size)
        if diff != type_size:
            return diff
    return 0


def region_swab(
    data: bytes,
    pe_list: list[EntryInfo],
    d_length: int,
    data_start: int,
    data_end: int,
) -> tuple[list[IndexEntry], int]:
    index_entries = []

    for index, pkg_entry in enumerate(pe_list):
        index_entry = IndexEntry(info=ei2h(pkg_entry), length=0, rd_len=0, data=b"")

        # Calculate start offset
        start = data_start + index_entry.info.offset
        if start >= data_end:
            error_msg = "Invalid data offset"
            raise ValueError(error_msg)

        # Calculate length for the current index entry
        if index < len(pe_list) - 1 and TYPE_SIZES[index_entry.info.type] == -1:
            index_entry.length = htonl(pe_list[index + 1].offset) - index_entry.info.offset
        else:
            index_entry.length = data_length(
                data,
                index_entry.info.type,
                index_entry.info.count,
                start,
                data_end,
            )

        if index_entry.length < 0:
            error_msg = "Invalid data length"
            raise ValueError(error_msg)

        # Calculate the end index
        end = start + index_entry.length
        if start > len(data) or end > len(data):
            error_msg = "Invalid data length"
            raise ValueError(error_msg)

        # Extract the data from the byte array
        index_entry.data = data[start:end]
        index_entries.append(index_entry)

        # Update data length (dl) with alignment
        d_length += index_entry.length + align_diff(index_entry.info.type, d_length)

    return index_entries, d_length


def hdrblob_verify_info(
    blob: HdrBlob,
    data: bytes,
) -> None:
    end = 0
    pe_offset = 0
    if blob.region_tag != 0:
        pe_offset = 1

    for pkg_entry in blob.pe_list[pe_offset:]:
        info = ei2h(pkg_entry)

        if end > info.offset:
            error_msg = f"Invalid offset info: {info}"
            raise ValueError(error_msg)

        if hdrchk_tag(info.tag):
            error_msg = f"Invalid tag info: {info}"
            raise ValueError(error_msg)

        if hdrchk_type(info.type):
            error_msg = f"Invalid type info: {info}"
            raise ValueError(error_msg)

        if hdrchk_align(info.type, info.offset):
            error_msg = f"invalid align info: {info}"
            raise ValueError(error_msg)

        if hdrchk_range(blob.d_length, info.offset):
            error_msg = f"Invalid range info: {info}"
            raise ValueError(error_msg)

        length = data_length(
            data,
            info.type,
            info.count,
            blob.data_start + info.offset,
            blob.data_end,
        )
        end = info.offset + length
        if hdrchk_range(blob.d_length, end) or length <= 0:
            error_msg = f"Invalid data length info: {info}"
            raise ValueError(error_msg)


def hdrblob_import(blob: HdrBlob, data: bytes) -> list[IndexEntry]:
    index_entries: list[IndexEntry] = []
    dribble_index_entries: list[IndexEntry] = []

    # Parse the first entry
    entry = ei2h(blob.pe_list[0])

    # Handle v3 legacy headers
    if entry.tag >= RPMTAG_HEADERI18NTABLE:
        # v3 header, create a legacy region entry
        index_entries, rdlen = region_swab(data, blob.pe_list, 0, blob.data_start, blob.data_end)
    else:
        # v4 header or "upgraded" v3 header with legacy region
        ril = blob.ril
        if entry.offset == 0:
            ril = blob.i_length

        # Process region entries
        index_entries, rdlen = region_swab(
            data,
            blob.pe_list[1:ril],
            0,
            blob.data_start,
            blob.data_end,
        )
        if rdlen < 0:
            error_msg = "Invalid region length"
            raise ValueError(error_msg)

        # Process dribble entries if present
        if blob.ril < len(blob.pe_list) - 1:
            dribble_index_entries, rdlen = region_swab(
                data,
                blob.pe_list[ril:],
                rdlen,
                blob.data_start,
                blob.data_end,
            )
            if rdlen < 0:
                error_msg = "Invalid length of dribble entries"
                raise ValueError(error_msg)

            # Merge index entries and dribble entries into a unique tag map
            uniq_tag_map = {
                index_entry.info.tag: index_entry
                for index_entry in index_entries + dribble_index_entries
            }

            # Collect unique index entries
            index_entries = list(uniq_tag_map.values())

        # Add REGION_TAG_COUNT to rdlen
        rdlen += REGION_TAG_COUNT

    # Check if the calculated length matches the data length
    if rdlen != blob.d_length:
        error_msg = (
            f"The calculated length ({rdlen}) is different from the data length ({blob.d_length})"
        )
        raise ValueError(error_msg)

    return index_entries


def hdrblob_verify_region(blob: HdrBlob, data: bytes) -> None:
    einfo = ei2h(blob.pe_list[0])
    region_tag = _get_region_tag(einfo)

    if einfo.tag != region_tag:
        return

    _validate_region_tag(einfo)
    _check_region_offset(blob, einfo)
    _parse_and_validate_trailer(blob, data, einfo, region_tag)
    _process_trailer_and_update_blob(blob, data)


def _get_region_tag(einfo: EntryInfo) -> int:
    if einfo.tag in {
        RPMTAG_HEADERIMAGE,
        RPMTAG_HEADERSIGNATURES,
        RPMTAG_HEADERIMMUTABLE,
    }:
        return einfo.tag
    return 0


def _validate_region_tag(einfo: EntryInfo) -> None:
    if not (einfo.type == REGION_TAG_TYPE and einfo.count == REGION_TAG_COUNT):
        error_msg = "Invalid region tag"
        raise ValueError(error_msg)


def _check_region_offset(blob: HdrBlob, einfo: EntryInfo) -> None:
    if hdrchk_range(blob.d_length, einfo.offset + REGION_TAG_COUNT):
        error_msg = "Invalid region offset"
        raise ValueError(error_msg)


def _parse_and_validate_trailer(
    blob: HdrBlob,
    data: bytes,
    einfo: EntryInfo,
    region_tag: int,
) -> None:
    region_end = blob.data_start + einfo.offset
    if region_end > len(data) or region_end + REGION_TAG_COUNT > len(data):
        error_msg = "Invalid region offset"
        raise ValueError(error_msg)

    trailer_data = data[region_end : region_end + REGION_TAG_COUNT]
    try:
        result = struct.unpack("<iIiI", trailer_data)
        EntryInfo(tag=result[0], type=result[1], offset=result[2], count=result[3])
    except struct.error as exc:
        error_msg = "Failed to parse trailer"
        raise ValueError(error_msg) from exc

    blob.rdl = region_end + REGION_TAG_COUNT - blob.data_start

    if region_tag == RPMTAG_HEADERSIGNATURES and einfo.tag == RPMTAG_HEADERIMAGE:
        einfo.tag = RPMTAG_HEADERSIGNATURES

    if not (
        einfo.tag == region_tag
        and einfo.type == REGION_TAG_TYPE
        and einfo.count == REGION_TAG_COUNT
    ):
        error_msg = "Invalid region trailer"
        raise ValueError(error_msg)


def _process_trailer_and_update_blob(blob: HdrBlob, data: bytes) -> None:
    result = struct.unpack(
        "<iIiI",
        data[blob.data_start + blob.rdl - REGION_TAG_COUNT : blob.data_start + blob.rdl],
    )
    trailer = EntryInfo(tag=result[0], type=result[1], offset=result[2], count=result[3])
    einfo = ei2h(trailer)
    einfo.offset = -einfo.offset
    blob.ril = einfo.offset // struct.calcsize("<iIiI")

    if (
        einfo.offset % REGION_TAG_COUNT != 0
        or hdrchk_range(blob.i_length, blob.ril)
        or hdrchk_range(blob.d_length, blob.rdl)
    ):
        error_msg = f"Invalid region size, region {blob.region_tag}"
        raise ValueError(error_msg)

    blob.region_tag = blob.pe_list[0].tag


def hdrblob_init(data: bytes) -> HdrBlob | None:
    try:
        blob = _create_hdrblob(data)
        _calculate_blob_fields(blob)
        _read_pe_list(blob, data)
        _verify_blob(blob, data)
        return blob
    except ValueError as exc:
        error_msg = "Failed to initialize header blob"
        raise ValueError(error_msg) from exc


def _create_hdrblob(data: bytes) -> HdrBlob:
    reader = struct.Struct(">I")
    (ideal_length,) = reader.unpack_from(data, 0)
    (d_length,) = reader.unpack_from(data, 4)
    return HdrBlob(
        pe_list=[],
        i_length=ideal_length,
        d_length=d_length,
        pv_len=0,
        data_start=0,
        data_end=0,
        region_tag=0,
        ril=0,
        rdl=0,
    )


def _calculate_blob_fields(blob: HdrBlob) -> None:
    entry_info_size = struct.calcsize("<iIiI")
    blob.data_start = 4 + 4 + blob.i_length * entry_info_size
    blob.pv_len = blob.data_start + blob.d_length
    blob.data_end = blob.data_start + blob.d_length
    if blob.i_length < 1:
        error_msg = "Region no tags error"
        raise ValueError(error_msg)


def _read_pe_list(blob: HdrBlob, data: bytes) -> None:
    entry_info_size = struct.calcsize("<iIiI")
    offset = 8
    for _ in range(blob.i_length):
        entry_info_data = data[offset : offset + entry_info_size]
        if len(entry_info_data) < entry_info_size:
            error_msg = "Failed to read entry info: incomplete data"
            raise ValueError(error_msg)
        pkg_entry = struct.unpack("<iIiI", entry_info_data)
        blob.pe_list.append(
            EntryInfo(
                tag=pkg_entry[0],
                type=pkg_entry[1],
                offset=pkg_entry[2],
                count=pkg_entry[3],
            ),
        )
        offset += entry_info_size
    if blob.pv_len >= HEADER_MAX_BYTES:
        error_msg = (
            f"Blob size({blob.pv_len}) BAD, 8 + 16 * il({blob.i_length}) + dl({blob.d_length})"
        )
        raise ValueError(error_msg)


def _verify_blob(blob: HdrBlob, data: bytes) -> None:
    hdrblob_verify_region(blob, data)
    hdrblob_verify_info(blob, data)


def header_import(data: bytes) -> list[IndexEntry] | None:
    def _validate_blob(blob: HdrBlob | None) -> HdrBlob:
        if blob is None:
            error_msg = "Failed to initialize header blob"
            raise ValueError(error_msg)
        return blob

    try:
        # Initialize the hdrblob from the data
        blob = _validate_blob(hdrblob_init(data))
        return hdrblob_import(blob, data)

    except ValueError as exc:
        error_msg = "Header import failed"
        raise ValueError(error_msg) from exc



import logging
from datetime import (
    datetime,
)

from packageurl import (
    PackageURL,
)
from pydantic import (
    BaseModel,
    ValidationError,
)

from labels.file.location import (
    Location,
)
from labels.linux.release import (
    Release,
)
from labels.model.core import (
    Digest,
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.common import (
    purl_qualifiers,
)
from labels.utils.licenses.validation import (
    validate_licenses,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


class AlpmFileRecord(BaseModel):
    path: str
    type: str | None = None
    uid: str | None = None
    gid: str | None = None
    time: datetime | None = None
    size: str | None = None
    link: str | None = None
    digests: list[Digest] | None = None


class AlpmDBEntry(BaseModel):
    licenses: str = ""
    base_package: str = ""
    package: str = ""
    version: str = ""
    description: str = ""
    architecture: str = ""
    size: int = 0
    packager: str = ""
    url: str = ""
    validation: str = ""
    reason: int = 0
    files: list[AlpmFileRecord] | None = None
    backup: list[AlpmFileRecord] | None = None

    def owned_files(self) -> list[str]:
        seen = set()
        result = []
        for file_record in self.files or []:
            if file_record.path and file_record.path not in seen:
                seen.add(file_record.path)
                result.append(file_record.path)
        result.sort()
        return result


def package_url(entry: AlpmDBEntry, distro: Release | None = None) -> str:
    qualifiers = {"arch": entry.architecture}
    if entry.base_package:
        qualifiers["upstream"] = entry.base_package
    return PackageURL(
        type="alpm",
        name=entry.package,
        version=entry.version,
        qualifiers=purl_qualifiers(qualifiers, distro),  # type: ignore
        subpath="",
    ).to_string()


def new_package(
    entry: AlpmDBEntry,
    release: Release | None,
    db_location: Location,
) -> Package | None:
    name = entry.package
    version = entry.version

    if not name or not version:
        return None

    licenses_candidates = entry.licenses.split("\n")

    try:
        return Package(
            name=name,
            version=version,
            locations=[db_location],
            licenses=validate_licenses(licenses_candidates),
            type=PackageType.AlpmPkg,
            metadata=entry,
            p_url=package_url(entry, release),
            language=Language.UNKNOWN_LANGUAGE,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": db_location.path(),
                },
            },
        )
        return None



import logging
from copy import (
    deepcopy,
)

from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.file.scope import Scope
from labels.internal.collection import (
    toml,
)
from labels.internal.collection.types import (
    IndexedDict,
    ParsedValue,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.python.model import (
    PythonPackage,
)
from labels.pkg.cataloger.python.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int, scope: Scope) -> Location:
    location.scope = scope
    location.dependency_type = DependencyType.DIRECT
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def _get_location_alt(location: Location, scope: Scope) -> Location:
    location.scope = scope
    location.dependency_type = DependencyType.DIRECT
    return location


def _get_packages(
    reader: LocationReadCloser,
    toml_packages: IndexedDict[str, ParsedValue],
    *,
    is_dev: bool = False,
) -> list[Package]:
    result = []
    scope = Scope.DEV if is_dev else Scope.PROD
    for package, version_data in toml_packages.items():
        version: str = ""
        if isinstance(version_data, str):
            version = version_data.strip("=<>~^ ")
        if isinstance(version_data, IndexedDict):
            version = str(version_data.get("version", "*")).strip("=<>~^ ")

        if not package or not version or "*" in version:
            continue
        current_location = deepcopy(reader.location)
        try:
            result.append(
                Package(
                    name=package,
                    version=version,
                    locations=[
                        (
                            _get_location(
                                current_location,
                                package.position.start.line,
                                scope,
                            )
                            if isinstance(package, IndexedDict)
                            else _get_location_alt(current_location, scope)
                        ),
                    ],
                    language=Language.PYTHON,
                    type=PackageType.PythonPkg,
                    metadata=PythonPackage(
                        name=package,
                        version=version,
                    ),
                    p_url=package_url(
                        name=package,
                        version=version,
                        package=None,
                    ),
                    licenses=[],
                    is_dev=is_dev,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": reader.location.path(),
                    },
                },
            )
            continue
    return result


def parse_pipfile_deps(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    file_content = reader.read_closer.read()
    toml_content: IndexedDict[str, ParsedValue] = toml.parse_toml_with_tree_sitter(file_content)
    toml_packages = toml_content.get("packages")
    if not isinstance(toml_packages, IndexedDict):
        return [], []
    packages = _get_packages(reader, toml_packages)
    dev_deps = toml_content.get("dev-packages")
    if isinstance(dev_deps, IndexedDict):
        dev_pkgs = _get_packages(reader, dev_deps, is_dev=True)
        packages.extend(dev_pkgs)
    return packages, []



import logging
from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from labels.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.internal.collection.types import (
    IndexedDict,
    IndexedList,
    ParsedValue,
)
from labels.internal.collection.yaml import (
    parse_yaml_with_tree_sitter,
)
from labels.model.core import (
    Package,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.swift.package import (
    new_cocoa_pods_package,
)

LOGGER = logging.getLogger(__name__)


def parse_podfile_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    try:
        podfile: IndexedDict[str, ParsedValue] = cast(
            IndexedDict[str, ParsedValue],
            parse_yaml_with_tree_sitter(reader.read_closer.read()),
        )
    except ValueError:
        return [], []

    if not podfile or "PODS" not in podfile:
        return [], []

    packages, dependencies_index = process_pods(podfile, reader)
    if not packages:
        return [], []

    relationships = generate_relations(dependencies_index, packages)

    return packages, relationships


def process_pods(
    podfile: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
) -> tuple[list[Package], dict[str, list[str]]]:
    packages: list[Package] = []
    dependencies_index: dict[str, list[str]] = {}

    direct_dependencies = podfile["DEPENDENCIES"]
    pods = podfile["PODS"]
    if not isinstance(direct_dependencies, IndexedList) or not isinstance(pods, IndexedList):
        return [], {}

    for index, pod in enumerate(pods):
        pod_name, pod_version = extract_pod_info(pod)
        if not pod_name or not pod_version:
            return [], {}

        pod_root_package = pod_name.split("/")[0]
        checksums = podfile["SPEC CHECKSUMS"]
        if not isinstance(checksums, IndexedDict):
            return [], {}

        new_location = deepcopy(reader.location)
        if new_location.coordinates:
            new_location.dependency_type = (
                DependencyType.DIRECT
                if pod_name in direct_dependencies
                else DependencyType.TRANSITIVE
            )
            new_location.coordinates.line = pods.get_position(index).start.line

        if pkg := new_cocoa_pods_package(
            pod_name,
            pod_version,
            checksums[pod_root_package],
            new_location,
        ):
            packages.append(pkg)
            dependencies_index[pod_name] = dependencies_index.get(pod_name, [])

    return packages, dependencies_index


def extract_pod_info(pod: ParsedValue) -> tuple[str, str]:
    if isinstance(pod, str | IndexedDict):
        pod_blob = pod if isinstance(pod, str) else next(iter(pod))
        pod_name = pod_blob.split(" ")[0]
        pod_version = pod_blob.split(" ")[1].strip("()")
        return pod_name, pod_version
    return "", ""


def generate_relations(
    dependencies_index: dict[str, list[str]],
    packages: list[Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []
    for package_name, dependencies in dependencies_index.items():
        package = next(x for x in packages if x.name == package_name)
        relationships.extend(
            Relationship(
                from_=package_dep,
                to_=package,
                type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
            )
            for dependency in dependencies
            if (
                package_dep := next(
                    (x for x in packages if x.name == dependency.split(" ")[0]),
                    None,
                )
            )
        )

    return relationships



import logging

from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

from labels.advisories import (
    roots as advisories,
)
from labels.file.location import (
    Location,
)
from labels.file.resolver import (
    Resolver,
)
from labels.internal.package_information.python import (
    PyPIResponse,
    get_pypi_package,
)
from labels.model.core import (
    Artifact,
    Digest,
    HealthMetadata,
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.common import (
    infer_algorithm,
)
from labels.pkg.cataloger.python.model import (
    PythonPackage,
)
from labels.pkg.cataloger.python.parse_wheel_egg_metadata import (
    ParsedData,
)
from labels.utils.licenses.validation import (
    validate_licenses,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def new_package_for_package(
    _resolver: Resolver,
    data: ParsedData,
    sources: Location,
) -> Package | None:
    name = data.python_package.name
    version = data.python_package.version

    if not name or not version:
        return None

    try:
        return Package(
            name=name,
            version=version,
            p_url=package_url(
                name,
                version,
                data.python_package,
            ),
            locations=[sources],
            language=Language.PYTHON,
            type=PackageType.PythonPkg,
            metadata=data.python_package,
            licenses=[],
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": sources.path(),
                },
            },
        )
        return None


def package_url(name: str, version: str, package: PythonPackage | None) -> str:
    return PackageURL(
        type="pypi",
        namespace="",
        name=name,
        version=version,
        qualifiers=_purl_qualifiers_for_package(package),
        subpath="",
    ).to_string()


def _purl_qualifiers_for_package(
    package: PythonPackage | None,
) -> dict[str, str]:
    if not package:
        return {}
    if (
        hasattr(package, "direct_url_origin")
        and package.direct_url_origin
        and package.direct_url_origin.vcs
    ):
        url = package.direct_url_origin
        return {"vcs_url": f"{url.vcs}+{url.url}@{url.commit_id}"}
    return {}


def _set_health_metadata(
    package: Package,
    pypi_package: PyPIResponse,
    current_package: PyPIResponse | None,
) -> None:
    info = pypi_package.get("info")
    if not isinstance(info, dict):
        return
    pypi_package_version = info.get("version")
    releases = pypi_package.get("releases")
    if not isinstance(releases, dict):
        releases = {}

    upload_time = releases.get(pypi_package_version) if pypi_package_version else []
    if not isinstance(upload_time, list):
        upload_time = []

    latest_version_created_at: str | None = None
    if upload_time:
        first_release = upload_time[0]
        time_value = first_release.get("upload_time_iso_8601")
        if isinstance(time_value, str):
            latest_version_created_at = time_value

    package.health_metadata = HealthMetadata(
        latest_version=pypi_package_version,
        latest_version_created_at=latest_version_created_at,
        authors=_get_authors(pypi_package),
        artifact=_get_artifact(package, current_package) if current_package else None,
    )


def _get_artifact(package: Package, current_package: PyPIResponse) -> Artifact | None:
    url = next(
        (x for x in current_package["urls"] if x["url"].endswith(".tar.gz")),
        None,
    )

    digest_value: str | None = url.get("digests", {}).get("sha256") or None if url else None

    return Artifact(
        url=url["url"] if url else f"https://pypi.org/pypi/{package.name}",
        integrity=Digest(
            algorithm=infer_algorithm(digest_value),
            value=digest_value,
        ),
    )


def _get_authors(pypi_package: PyPIResponse) -> str | None:
    package_info = pypi_package["info"]
    author: str | None = None
    package_author = package_info["author"]
    author_email = package_info.get("author_email")
    if package_author:
        author = package_author
    if not author and author_email:
        author = author_email
    if author and author_email and author_email not in author:
        author = f"{author} <{author_email}>"
    return author


def _update_licenses(pypi_package: PyPIResponse, package: Package) -> None:
    info = pypi_package.get("info")
    if not isinstance(info, dict):
        return
    licenses = info.get("license")
    if licenses:
        package.licenses = validate_licenses([licenses])


def complete_package(package: Package) -> Package:
    pkg_advisories = advisories.get_vulnerabilities("pip", package.name, package.version)
    if pkg_advisories:
        package.advisories = pkg_advisories

    pypi_package = get_pypi_package(package.name)
    if not pypi_package:
        return package

    current_package = get_pypi_package(package.name, package.version)

    _set_health_metadata(package, pypi_package, current_package)

    _update_licenses(pypi_package, package)

    return package



import logging

from packageurl import (
    PackageURL,
)
from pydantic import (
    BaseModel,
    ValidationError,
)

from labels.advisories import (
    roots as advisories,
)
from labels.file.location import (
    Location,
)
from labels.internal.package_information.elixir import (
    get_hex_package,
)
from labels.model.core import (
    HealthMetadata,
    Language,
    Package,
    PackageType,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


class ElixirMixLockEntry(BaseModel):
    name: str
    version: str
    pkg_hash: str
    pkg_hash_ext: str


def new_package(entry: ElixirMixLockEntry, locations: Location) -> Package | None:
    name = entry.name
    version = entry.version

    if not name or not version:
        return None

    try:
        return Package(
            name=name,
            version=version,
            type=PackageType.HexPkg,
            locations=[locations],
            p_url=package_url(name, version),
            metadata=entry,
            language=Language.ELIXIR,
            licenses=[],
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": locations.path(),
                },
            },
        )
        return None


def package_url(name: str, version: str) -> str:
    return PackageURL(  # type: ignore
        type="hex",
        namespace="",
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def _update_advisories(package: Package) -> None:
    pkg_advisories = advisories.get_vulnerabilities("erlang", package.name, package.version)
    if pkg_advisories:
        package.advisories = pkg_advisories


def complete_package(package: Package) -> Package:
    _update_advisories(package)
    response = get_hex_package(package.name)
    if not response:
        return package
    package.health_metadata = HealthMetadata(
        latest_version=response["latest_stable_version"],
        latest_version_created_at=next(
            x["version"]
            for x in response["releases"]
            if x["version"] == response["latest_stable_version"]
        ),
    )
    if not package.licenses:
        package.licenses = response["meta"]["licenses"]
    if response["owners"]:
        package.health_metadata.authors = ", ".join([x["username"] for x in response["owners"]])
    return package



import logging
from copy import deepcopy
from typing import TYPE_CHECKING, cast

from packageurl import PackageURL
from pydantic import BaseModel, ValidationError
from urllib3.util import parse_url

from labels.artifact.relationship import Relationship
from labels.file.dependency_type import DependencyType
from labels.utils.strings import format_exception

if TYPE_CHECKING:
    from labels.file.location import Location
from labels.file.location_read_closer import LocationReadCloser
from labels.file.resolver import Resolver
from labels.internal.collection.types import IndexedDict, Position
from labels.internal.collection.yaml import parse_yaml_with_tree_sitter
from labels.model.core import Language, Package, PackageType
from labels.pkg.cataloger.generic.parser import Environment


class DartPubspecLickEntry(BaseModel):
    name: str
    version: str
    hosted_url: str
    vcs_url: str


LOGGER = logging.getLogger(__name__)


def get_hosted_url(entry: IndexedDict[str, str]) -> str:
    hosted: str | None = entry.get("hosted")
    description: dict[str, str] | None = cast(
        dict[str, str] | None,
        entry.get("description"),
    )

    if hosted == "hosted" and description and description["url"] != "https://pub.dartlang.org":
        if host := parse_url(description["url"]).host:
            return host

        return description["url"]

    return ""


def get_vcs_url(entry: IndexedDict[str, str]) -> str:
    source: str | None = entry.get("source")
    description: dict[str, str] | None = cast(
        dict[str, str] | None,
        entry.get("description"),
    )

    if description and source == "git":
        if description.get("path") == ".":
            return f'{description["url"]}@{description["resolved-ref"]}'
        return description["url"] + f'@{description["resolved-ref"]}' + f'#{description["path"]}'
    return ""


def package_url(entry: DartPubspecLickEntry) -> str:
    qualifiers = {}
    if entry.hosted_url:
        qualifiers["hosted_url"] = entry.hosted_url
    elif entry.vcs_url:
        qualifiers["vcs_url"] = entry.vcs_url

    return PackageURL(  # type: ignore
        type="pub",
        namespace="",
        name=entry.name,
        version=entry.version,
        qualifiers=qualifiers,
        subpath="",
    ).to_string()


def parse_pubspec_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_yaml = cast(
        IndexedDict[str, IndexedDict[str, IndexedDict[str, str]]],
        parse_yaml_with_tree_sitter(reader.read_closer.read()),
    )
    packages: list[Package] = []
    relationships: list[Relationship] = []
    yaml_packages: IndexedDict[str, IndexedDict[str, str]] = package_yaml["packages"]
    items = yaml_packages.items()

    for package_name, package_value in items:
        is_transitive = package_value.get("dependency") == "transitive"
        version: str | None = package_value.get("version")

        if not package_name or not version:
            continue

        current_location: Location = deepcopy(reader.location)
        if current_location.coordinates:
            position: Position = yaml_packages.get_key_position(package_name)
            current_location.coordinates.line = position.start.line
            current_location.dependency_type = (
                DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
            )
        metadata = DartPubspecLickEntry(
            name=package_name,
            version=version,
            hosted_url=get_hosted_url(package_value),
            vcs_url=get_vcs_url(package_value),
        )
        try:
            packages.append(
                Package(
                    name=package_name,
                    version=version,
                    locations=[current_location],
                    language=Language.DART,
                    licenses=[],
                    type=PackageType.DartPubPkg,
                    p_url=package_url(metadata),
                    metadata=metadata,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": current_location.path(),
                    },
                },
            )
            continue

    return packages, relationships



from typing import NotRequired, TypedDict, cast

import requests

from labels.internal.cache import dual_cache


class RubyGemsMetadata(TypedDict, total=False):
    homepage_uri: str
    changelog_uri: str
    bug_tracker_uri: str
    source_code_uri: str
    documentation_uri: str


class RubyGemsDependencies(TypedDict):
    development: list[str]
    runtime: list[str]


class RubyGemsPackage(TypedDict, total=False):
    name: str
    downloads: int
    version: str
    version_created_at: str
    version_downloads: int
    platform: str
    authors: str
    info: str
    licenses: list[str]
    metadata: RubyGemsMetadata
    yanked: bool
    sha: str
    spec_sha: str
    project_uri: str
    gem_uri: str
    homepage_uri: str
    wiki_uri: NotRequired[str]
    documentation_uri: str
    mailing_list_uri: NotRequired[str]
    source_code_uri: str
    bug_tracker_uri: str
    changelog_uri: str
    funding_uri: NotRequired[str]
    dependencies: RubyGemsDependencies
    built_at: NotRequired[str]
    created_at: NotRequired[str]
    description: NotRequired[str]
    downloads_count: NotRequired[int]
    number: NotRequired[str]
    summary: NotRequired[str]
    rubygems_version: NotRequired[str]
    ruby_version: NotRequired[str]
    prerelease: NotRequired[bool]
    requirements: NotRequired[list[str]]


@dual_cache
def get_gem_package(package_name: str, version: str | None = None) -> RubyGemsPackage | None:
    if version:
        url = f"https://rubygems.org/api/v2/rubygems/{package_name}/versions/{version}.json"
    else:
        url = f"https://rubygems.org/api/v1/gems/{package_name}.json"

    response = requests.get(url, timeout=30)
    if response.status_code == 200:
        return cast(RubyGemsPackage, response.json())
    return None



import logging
import re
from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.file.scope import (
    Scope,
)
from labels.internal.collection.types import (
    IndexedDict,
    IndexedList,
    ParsedValue,
)
from labels.internal.collection.yaml import (
    parse_yaml_with_tree_sitter,
)
from labels.model.core import (
    Digest,
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.javascript.model import (
    PnpmEntry,
)
from labels.pkg.cataloger.javascript.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)

VERSION_PATTERN = re.compile(r"(\d+\.\d+\.\d+(-[0-9A-Za-z\.]+)?)")


def extract_package_name_from_key_dependency(item: str) -> str | None:
    # Regex pattern to extract the package name
    pattern = r"^@?[\w-]+/[\w-]+$"
    match = re.match(pattern, item)
    if match:
        return match.group(0)
    return None


def extract_version_from_value_dependency(item: str) -> str | None:
    # Regex pattern to extract the version number before any parentheses
    pattern = r"^(\d+\.\d+\.\d+)"
    match = re.match(pattern, item)
    if match:
        return match.group(1)
    return None


def _get_package(
    packages: list[Package],
    dep_name: str | None,
    dep_version: str | None,
) -> Package | None:
    return next(
        (x for x in packages if x.name == dep_name and x.version == dep_version),
        None,
    )


def _process_relationships(
    dependencies: IndexedDict[str, ParsedValue],
    packages: list[Package],
    current_package: Package | None,
) -> list[Relationship]:
    relationships: list[Relationship] = []
    for raw_dep_name, raw_dep_version in dependencies.items():
        if not isinstance(raw_dep_version, str):
            continue
        dep_name = extract_package_name_from_key_dependency(
            raw_dep_name,
        )
        dep_version = extract_version_from_value_dependency(
            raw_dep_version,
        )
        if (
            dep := _get_package(
                packages,
                dep_name,
                dep_version,
            )
            and current_package
        ):
            relationships.append(
                Relationship(
                    from_=dep,
                    to_=current_package,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                ),
            )
    return relationships


def _generate_relations_relationship(
    package_yaml: IndexedDict[str, ParsedValue],
    packages: list[Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []
    packages_items = package_yaml["packages"]
    if not isinstance(packages_items, IndexedDict):
        return relationships
    for package_key, package_value in packages_items.items():
        if not isinstance(package_value, IndexedDict):
            continue
        if match_ := re.search(r"/(@?[^@]+)@(\d+\.\d+\.\d+)", package_key):
            package_name = match_.groups()[0]
            package_version = match_.groups()[1]
            current_package = _get_package(
                packages,
                dep_name=package_name,
                dep_version=package_version,
            )
            dependencies = package_value.get("dependencies")
            if dependencies and isinstance(dependencies, IndexedDict):
                relationships.extend(
                    _process_relationships(dependencies, packages, current_package),
                )
    return relationships


def _get_package_metadata(
    package_value: IndexedDict[str, ParsedValue],
    *,
    is_dev: bool,
) -> PnpmEntry:
    resolution_value = package_value.get("resolution")
    integrity_value = (
        resolution_value.get("integrity") if isinstance(resolution_value, IndexedDict) else None
    )

    integrity_value = integrity_value if isinstance(integrity_value, str) else None

    return PnpmEntry(
        is_dev=is_dev if isinstance(is_dev, bool) else False,
        integrity=Digest(
            algorithm="sha-512",
            value=integrity_value,
        ),
    )


def manage_coordinates(
    package_yaml: IndexedDict[str, ParsedValue],
    package_key: str,
    package_name: str,
    direct_dependencies: list[ParsedValue],
    base_location: Location,
) -> Location:
    current_location = deepcopy(base_location)
    if current_location.coordinates:
        packages = package_yaml["packages"]
        if not isinstance(packages, IndexedDict):
            return current_location
        position = packages.get_key_position(package_key)
        current_location.coordinates.line = position.start.line
        is_transitive = package_name not in direct_dependencies
        current_location.dependency_type = (
            DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
        )
    return current_location


def process_package_string(
    package: str,
    spec: IndexedDict[str, ParsedValue],
) -> tuple[str, str] | None:
    if package.startswith("github"):
        pkg_name = spec.get("name", "")
        pkg_version = spec.get("version", "")
    else:
        pkg_info: list[str] = VERSION_PATTERN.split(package.strip("\"'"))
        if len(pkg_info) < 2:
            return None
        pkg_name = pkg_info[0].lstrip("/")[0:-1]
        pkg_version = pkg_info[1]

    if not isinstance(pkg_name, str) or not isinstance(pkg_version, str):
        return None
    return pkg_name, pkg_version


def _process_package(
    package_key: str,
    pkg_spec: IndexedDict[str, ParsedValue],
    package_yaml: IndexedDict[str, ParsedValue],
    direct_dependencies: list[ParsedValue],
    base_location: Location,
) -> Package | None:
    if match_ := process_package_string(package_key, pkg_spec):
        package_name = match_[0]
        package_version = match_[1]

        if not package_name or not package_version:
            return None
        is_dev = pkg_spec.get("dev")
        current_location: Location = manage_coordinates(
            package_yaml,
            package_key,
            package_name,
            direct_dependencies,
            base_location,
        )
        is_dev = is_dev if isinstance(is_dev, bool) else False
        current_location.scope = Scope.DEV if is_dev else Scope.PROD
        try:
            return Package(
                name=package_name,
                version=package_version,
                locations=[current_location],
                language=Language.JAVASCRIPT,
                licenses=[],
                type=PackageType.NpmPkg,
                p_url=package_url(package_name, package_version),
                metadata=_get_package_metadata(pkg_spec, is_dev=is_dev),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": current_location.path(),
                    },
                },
            )
    return None


def parse_pnpm_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_yaml: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        parse_yaml_with_tree_sitter(reader.read_closer.read()),
    )

    if not package_yaml:
        return [], []
    dependencies = package_yaml.get("dependencies")
    dev_dependencies = package_yaml.get("devDependencies")
    direct_dependencies: list[ParsedValue] = []
    if isinstance(dependencies, IndexedList) and isinstance(dev_dependencies, IndexedList):
        direct_dependencies = [*dev_dependencies, *dependencies]
    packages: list[Package] = []
    relationships: list[Relationship] = []
    packages_items = package_yaml["packages"]
    if not isinstance(packages_items, IndexedDict):
        return [], []
    for package_key, pkg_spec in packages_items.items():
        if not isinstance(pkg_spec, IndexedDict):
            continue
        if package := _process_package(
            package_key,
            pkg_spec,
            package_yaml,
            direct_dependencies,
            reader.location,
        ):
            packages.append(package)

    relationships = _generate_relations_relationship(package_yaml, packages)

    return packages, relationships



from pydantic import (
    BaseModel,
)

from labels.model.core import (
    Digest,
)


class NpmPackage(BaseModel):
    name: str
    version: str | None = None
    author: str | None = None
    homepage: str | None = None
    description: str | None = None
    url: str | None = None
    private: bool | None = None
    is_dev: bool = False


class NpmPackageLockEntry(BaseModel):
    resolved: str | None = None
    integrity: str | None = None
    is_dev: bool = False


class YarnLockEntry(BaseModel):
    resolved: str | None = None
    integrity: str | None = None


class PnpmEntry(BaseModel):
    is_dev: bool = False
    integrity: Digest | None = None



import logging
from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from labels.internal.collection.types import (
    IndexedDict,
    IndexedList,
    ParsedValue,
)
from labels.model.core import (
    Package,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.swift.package import (
    new_swift_package_manager_package,
)

LOGGER = logging.getLogger(__name__)
EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()
EMPTY_LIST: IndexedList[ParsedValue] = IndexedList()


def _get_name_and_version(
    pin: ParsedValue,
) -> tuple[str, str, IndexedDict[str, ParsedValue]] | None:
    if not isinstance(pin, IndexedDict):
        return None
    state: ParsedValue = pin.get("state", EMPTY_DICT)
    name = pin.get("identity")
    if not isinstance(state, IndexedDict):
        return None
    version = state.get("version")

    if not name or not version or not isinstance(version, str) or not isinstance(name, str):
        return None

    return name, version, state


def parse_package_resolved(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_resolved: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )

    packages: list[Package] = []
    relationships: list[Relationship] = []
    package_resolved_pins: ParsedValue = package_resolved.get("pins", EMPTY_LIST)

    if isinstance(package_resolved_pins, IndexedList):
        for pin in package_resolved_pins:
            if not isinstance(pin, IndexedDict):
                continue
            info = _get_name_and_version(pin)
            if not info:
                continue
            new_location = deepcopy(reader.location)
            if new_location.coordinates:
                new_location.coordinates.line = pin.position.start.line

            if pkg := new_swift_package_manager_package(
                name=info[0],
                version=info[1],
                source_url=pin.get("location"),
                revision=info[2].get("revision"),
                location=new_location,
            ):
                packages.append(pkg)

    return packages, relationships



import logging
from typing import TYPE_CHECKING

from labels.utils.strings import format_exception

if TYPE_CHECKING:
    from collections.abc import ItemsView

from pydantic import ValidationError

from labels.artifact.relationship import Relationship
from labels.file.dependency_type import DependencyType
from labels.file.location import Location
from labels.file.location_read_closer import LocationReadCloser
from labels.file.resolver import Resolver
from labels.file.scope import Scope
from labels.internal.collection.toml import parse_toml_with_tree_sitter
from labels.internal.collection.types import IndexedDict, ParsedValue
from labels.model.core import Language, Package, PackageType
from labels.pkg.cataloger.generic.parser import Environment
from labels.pkg.cataloger.rust.package import package_url

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        location.dependency_type = DependencyType.DIRECT
        return location.model_copy(update=l_upd)
    return location


def _get_version(value: ParsedValue) -> str | None:
    if isinstance(value, str):
        return value
    if not isinstance(value, IndexedDict):
        return None
    if "git" in value:
        repo_url: str = str(value.get("git", ""))
        branch: str = str(value.get("branch", ""))
        if repo_url and branch:
            return f"{repo_url}@{branch}"
    version: str = str(value.get("version", ""))
    return version


def _get_packages(
    reader: LocationReadCloser,
    dependencies: ParsedValue,
    *,
    is_dev: bool,
) -> list[Package]:
    if dependencies is None or not isinstance(dependencies, IndexedDict):
        return []

    packages = []

    general_location = _get_location(
        reader.location,
        dependencies.position.start.line,
    )
    items: ItemsView[str, ParsedValue] = dependencies.items()

    for name, value in items:
        version = _get_version(value)
        if not name or not version:
            continue

        location = (
            _get_location(reader.location, value.position.start.line)
            if isinstance(value, IndexedDict)
            else general_location
        )
        location.scope = Scope.DEV if is_dev else Scope.PROD
        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[location],
                    language=Language.RUST,
                    licenses=[],
                    p_url=package_url(name=name, version=version),
                    type=PackageType.RustPkg,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages


def parse_cargo_toml(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content: IndexedDict[
        str,
        ParsedValue,
    ] = parse_toml_with_tree_sitter(reader.read_closer.read())

    deps: ParsedValue = content.get(
        "dependencies",
    )
    dev_deps: ParsedValue = content.get(
        "dev-dependencies",
    )
    packages = [
        *_get_packages(reader, deps, is_dev=False),
        *_get_packages(reader, dev_deps, is_dev=True),
    ]
    return packages, []



import json
import logging
from datetime import UTC, datetime
from enum import Enum
from pathlib import Path
from typing import Any, cast

from jsonschema import validate
from jsonschema.exceptions import ValidationError
from pydantic import BaseModel

from labels.artifact.relationship import Relationship
from labels.file.resolver import Resolver
from labels.format.common import set_namespace_version
from labels.format.types import FLUID_SBOM_JSON_SCHEMA
from labels.internal.collection.types import IndexedDict, IndexedList
from labels.model.core import Package, SbomConfig
from labels.utils.exceptions import FluidJSONValidationError

LOGGER = logging.getLogger(__name__)


class EnumEncoder(json.JSONEncoder):
    def default(self, item: Any) -> Any:
        if isinstance(item, Enum):
            return item.value
        if isinstance(item, IndexedList):
            return list(item.data)

        if isinstance(item, IndexedDict):
            return dict(item.data)
        if isinstance(item, BaseModel):
            item = item.model_dump if item else None
        if isinstance(item, datetime):
            return item.isoformat()
        return json.JSONEncoder.default(self, item)


def validate_pkgs(sbom_pkgs: list[dict[str, Any]]) -> list[dict[str, Any]]:
    pkgs_to_remove = []
    for index, raw_pkg in enumerate(sbom_pkgs):
        try:
            pkg = json.loads(json.dumps(raw_pkg, cls=EnumEncoder))
            sbom_pkgs[index] = pkg
        except TypeError:
            pkgs_to_remove.append(index)

    for index in sorted(pkgs_to_remove, reverse=True):
        sbom_pkgs.pop(index)
    return sbom_pkgs


def format_fluid_sbom(
    *,
    packages: list[Package],
    relationships: list[Relationship],
    config: SbomConfig,
    resolver: Resolver,
) -> None:
    now_utc = datetime.now(UTC).isoformat()
    namespace, version = set_namespace_version(config=config, resolver=resolver)
    file_path = Path(f"{config.output}.json")
    sbom_pkgs = [
        {
            "id": package.id_,
            "name": package.name,
            "version": package.version,
            "locations": [
                {
                    "path": location.path(),
                    "line": location.coordinates.line
                    if location.coordinates and location.coordinates.line
                    else None,
                    "layer": location.coordinates.file_system_id
                    if location.coordinates and location.coordinates.file_system_id
                    else None,
                    "dependency_type": location.dependency_type.value,
                    "scope": location.scope.value,
                }
                for location in package.locations
            ],
            "licenses": package.licenses,
            "type": package.type.value,
            "language": package.language.value,
            "platform": package.language.get_platform_value(),
            "package_url": package.p_url,
            "found_by": package.found_by,
            "health_metadata": package.health_metadata.model_dump()
            if package.health_metadata
            else None,
            "advisories": [advisory.model_dump() for advisory in package.advisories or []],
        }
        for package in packages
    ]

    sbom_pkgs = validate_pkgs(sbom_pkgs)
    sbom_details = {
        "name": namespace,
        "version": version if version else None,
        "timestamp": now_utc,
        "tool": "Fluid-Labels",
        "organization": "Fluid attacks",
    }
    dependency_map: dict[str, list[str]] = {}

    for relationship in relationships:
        from_pkg = cast(Package, relationship.from_)
        to_pkg = cast(Package, relationship.to_)
        ref = f"{to_pkg.id_}"
        dep = f"{from_pkg.id_}"

        if ref not in dependency_map:
            dependency_map[ref] = []

        dependency_map[ref].append(dep)

    sbom_relationships = [
        {"from": ref, "to": depends_on_list} for ref, depends_on_list in dependency_map.items()
    ]
    result = {
        "sbom_details": sbom_details,
        "packages": sbom_pkgs,
        "relationships": sbom_relationships,
    }

    try:
        validate(result, FLUID_SBOM_JSON_SCHEMA)
    except ValidationError as ex:
        raise FluidJSONValidationError(ex) from None

    LOGGER.info(
        "🆗 Valid Fluid JSON format, generating output file at %s",
        str(file_path),
    )
    with file_path.open("w", encoding="utf-8") as json_file:
        json.dump(result, json_file, indent=4, cls=EnumEncoder)
    LOGGER.info("✅ Output file successfully generated")



import atexit
import logging
import os
import sqlite3
import subprocess  # nosec
from pathlib import Path
from typing import (
    Literal,
    cast,
)

import boto3
from botocore import (
    UNSIGNED,
)
from botocore.config import (
    Config,
)
from platformdirs import (
    user_data_dir,
)

from labels.advisories.roots.match_versions import match_vulnerable_versions
from labels.model.core import (
    Advisory,
)

LOGGER = logging.getLogger(__name__)

BUCKET_NAME = "skims.sca"
DB_NAME = "skims_sca_advisories.db"
BUCKET_FILE_KEY = f"{DB_NAME}.zst"
CONFIG_DIRECTORY = user_data_dir(
    appname="fluid-labels",
    appauthor="fluidattacks",
    ensure_exists=True,
)
DB_LOCAL_PATH = os.path.join(CONFIG_DIRECTORY, DB_NAME)
DB_LOCAL_COMPRESSED_PATH = f"{DB_LOCAL_PATH}.zst"
S3_SERVICE_NAME: Literal["s3"] = "s3"
S3_CLIENT = boto3.client(
    service_name=S3_SERVICE_NAME,
    config=Config(
        region_name="us-east-1",
        signature_version=UNSIGNED,
    ),
)


def get_package_advisories(
    package_manager: str,
    package_name: str,
) -> list[Advisory]:
    connection = DATABASE.get_connection()
    cursor = connection.cursor()
    cursor.execute(
        """
        SELECT
            adv_id,
            source,
            vulnerable_version,
            severity_level,
            severity,
            severity_v4,
            epss,
            details,
            percentile
        FROM advisories
        WHERE package_manager = ? AND package_name = ?;
        """,
        (package_manager, package_name),
    )
    return [
        Advisory(
            id=result[0],
            urls=[result[1]],
            version_constraint=result[2] or None,
            severity=result[3],
            cvss3=result[4],
            cvss4=result[5],
            epss=result[6] or 0.0,
            description=result[7] or None,
            percentile=result[8] or 0.0,
            cpes=[],
            namespace=package_manager,
        )
        for result in cursor.fetchall()
    ]


def get_vulnerabilities(platform: str, product: str, version: str) -> list[Advisory]:
    vulnerabilities = []
    if product and version and (advisories := get_package_advisories(platform, product.lower())):
        vulnerabilities = [
            advisor
            for advisor in advisories
            if match_vulnerable_versions(version.lower(), advisor.version_constraint)
        ]

    return vulnerabilities


def _get_database_file() -> None:
    LOGGER.info("⬇️ Downloading advisories database")
    S3_CLIENT.download_file(
        Bucket=BUCKET_NAME,
        Key=BUCKET_FILE_KEY,
        Filename=DB_LOCAL_COMPRESSED_PATH,
    )
    LOGGER.info("🗜️ Decompressing advisories database")
    with subprocess.Popen(  # noqa: S603
        ["zstd", "-d", "-f", DB_LOCAL_COMPRESSED_PATH],  # noqa: S607
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as process:
        _, stderr = process.communicate()
        if cast(int, process.returncode) != 0:
            raise RuntimeError(stderr.decode())


def initialize_db() -> bool:
    local_database_exists = Path(DB_LOCAL_PATH).is_file()

    try:
        db_metadata = S3_CLIENT.head_object(Bucket=BUCKET_NAME, Key=BUCKET_FILE_KEY)
        up_to_date = (
            local_database_exists
            and Path(DB_LOCAL_PATH).stat().st_mtime >= db_metadata["LastModified"].timestamp()
        )

        if up_to_date:
            LOGGER.info("✅ Advisories database is up to date")
            return True
        _get_database_file()
        Path(DB_LOCAL_COMPRESSED_PATH).unlink()
    except Exception:
        if local_database_exists:
            LOGGER.warning(
                "⚠️ Advisories may be outdated, unable to update database",
            )
            return True

        LOGGER.exception(
            "❌ Advisories won't be included, unable to download database",
        )
        return False
    else:
        return True


class Database:
    def __init__(self) -> None:
        self.connection: sqlite3.Connection | None = None

    def initialize(self) -> None:
        if self.connection is None and initialize_db():
            self.connection = sqlite3.connect(
                DB_LOCAL_PATH,
                check_same_thread=False,
            )
            atexit.register(self.connection.close)

    def get_connection(self) -> sqlite3.Connection:
        if self.connection is not None:
            return self.connection
        self.connection = sqlite3.connect(
            DB_LOCAL_PATH,
            check_same_thread=False,
        )
        atexit.register(self.connection.close)
        return self.connection


DATABASE = Database()



import logging
import re
from copy import (
    deepcopy,
)

from bs4 import (
    BeautifulSoup,
)
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)
PACKAGE = re.compile(r".+\\packages\\(?P<package_info>[^\s\\]*)\\.+")
DEP_INFO = re.compile(r"(?P<package_name>[^\d]*)\.(?P<version>[\d.]+)")


def _get_version(include_info: list[str]) -> str | None:
    return next(
        (
            pkg_info.lstrip("Version=")
            for pkg_info in include_info
            if pkg_info.startswith("Version=")
        ),
        None,
    )


def format_package(
    name: str,
    version: str,
    line: int,
    reader: LocationReadCloser,
) -> Package | None:
    location = deepcopy(reader.location)
    if location.coordinates:
        location.coordinates.line = line
        location.dependency_type = DependencyType.DIRECT

    try:
        return Package(
            name=name,
            version=version,
            locations=[location],
            language=Language.DOTNET,
            licenses=[],
            type=PackageType.DotnetPkg,
            metadata=None,
            p_url=PackageURL(
                type="nuget",
                namespace="",
                name=name,
                version=version,
                qualifiers={},
                subpath="",
            ).to_string(),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": location.path(),
                },
            },
        )
    return None


def _format_csproj_reference_deps(root: BeautifulSoup, reader: LocationReadCloser) -> list[Package]:
    return [
        package
        for pkg in root.find_all("reference", recursive=True)
        if (
            (dll_path := pkg.find("hintpath"))
            and (raw_package := PACKAGE.match(dll_path.text))
            and (pkg_info := DEP_INFO.match(raw_package.group("package_info")))
            and (
                package := format_package(
                    str(pkg_info.group("package_name").lower()),
                    str(pkg_info.group("version")),
                    dll_path.sourceline,
                    reader,
                )
            )
        )
        or (
            (include := pkg.get("include"))
            and (include_info := include.replace(" ", "").split(","))
            and (version := _get_version(include_info))
            and (
                package := format_package(
                    str(include_info[0]).strip(),
                    str(version),
                    pkg.sourceline,
                    reader,
                )
            )
        )
    ]


def parse_csproj(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    root = BeautifulSoup(reader.read_closer.read(), features="html.parser")
    for pkg in root.find_all("packagereference", recursive=True):
        name: str | None = pkg.get("include")
        version: str | None = pkg.get("version")
        if not name or not version:
            continue

        if package := format_package(name, version, pkg.sourceline, reader):
            packages.append(package)

    packages.extend(_format_csproj_reference_deps(root, reader))
    return packages, []



import logging
from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

from labels.advisories import (
    roots as advisories,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location import (
    Location,
)
from labels.file.scope import (
    Scope,
)
from labels.internal.collection.types import (
    IndexedDict,
    IndexedList,
    ParsedValue,
)
from labels.internal.package_information.php import (
    PackagistPackageInfo,
    get_composer_package,
)
from labels.model.core import (
    Artifact,
    Digest,
    HealthMetadata,
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.common import (
    infer_algorithm,
)
from labels.pkg.cataloger.php.model import (
    PhpComposerAuthors,
    PhpComposerExternalReference,
    PhpComposerInstalledEntry,
)
from labels.utils.licenses.validation import (
    validate_licenses,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)

EMPTY_LIST: IndexedList[str] = IndexedList()


def package_url(name: str, version: str) -> str:
    fields = name.split("/")

    vendor = ""
    if len(fields) == 1:
        name = fields[0]
    elif len(fields) >= 2:
        vendor = fields[0]
        name = "-".join(fields[1:])

    return PackageURL(  # type: ignore
        type="composer",
        namespace=vendor,
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def new_package_from_composer(
    package: IndexedDict[str, ParsedValue],
    location: Location,
    *,
    is_dev: bool = False,
) -> Package | None:
    empty_list_dict: IndexedList[IndexedDict[str, str]] = IndexedList()
    new_location = deepcopy(location)
    new_location.scope = Scope.DEV if is_dev else Scope.PROD
    if new_location.coordinates:
        new_location.coordinates.line = package.position.start.line
        new_location.dependency_type = DependencyType.DIRECT

    try:
        source = cast(IndexedDict[str, str], package.get("source"))
        dist = cast(IndexedDict[str, str], package.get("dist"))
        name = cast(str, package.get("name"))
        version = cast(str, package.get("version"))

        if not name or not version:
            return None

        return Package(
            name=name,
            version=version,
            locations=[new_location],
            language=Language.PHP,
            licenses=list(cast(IndexedList[str], package.get("license", EMPTY_LIST))),
            type=PackageType.PhpComposerPkg,
            p_url=package_url(name, version),
            metadata=PhpComposerInstalledEntry(
                name=name,
                version=version,
                source=PhpComposerExternalReference(
                    type=source.get("type"),
                    url=source.get("url"),
                    reference=source.get("reference"),
                    shasum=source.get("shasum"),
                )
                if source
                else None,
                dist=PhpComposerExternalReference(
                    type=dist.get("type"),
                    url=dist.get("url"),
                    reference=dist.get("reference"),
                    shasum=dist.get("shasum"),
                )
                if dist
                else None,
                require=cast(dict[str, str], package.get("require"))
                if isinstance(package.get("require"), IndexedDict)
                else None,
                provide=cast(dict[str, str], package.get("provide"))
                if isinstance(package.get("provide"), IndexedDict)
                else None,
                require_dev=cast(dict[str, str], package.get("require-dev"))
                if isinstance(package.get("require-dev"), IndexedDict)
                else None,
                suggest=cast(dict[str, str], package.get("suggest"))
                if isinstance(package.get("suggest"), IndexedDict)
                else None,
                license=cast(list[str], package.get("license"))
                if isinstance(package.get("license"), IndexedList)
                else None,
                type=cast(str, package.get("type"))
                if isinstance(package.get("type"), str)
                else None,
                notification_url=cast(str, package.get("notification-url"))
                if isinstance(package.get("notification-url"), str)
                else None,
                bin=cast(list[str], package.get("bin"))
                if isinstance(package.get("bin"), IndexedList)
                else None,
                authors=[
                    PhpComposerAuthors(
                        name=cast(str, x.get("name")),
                        email=x.get("email"),
                        homepage=x.get("homepage"),
                    )
                    for x in cast(
                        list[IndexedDict[str, str]],
                        package.get("authors", empty_list_dict),
                    )
                ],
                description=cast(str, package.get("description"))
                if isinstance(package.get("description"), str)
                else None,
                homepage=cast(str, package.get("homepage"))
                if isinstance(package.get("homepage"), str)
                else None,
                keywords=cast(list[str], package.get("keywords"))
                if isinstance(package.get("keywords"), IndexedDict)
                else None,
                time=cast(str, package.get("time"))
                if isinstance(package.get("time"), str)
                else None,
            ),
            is_dev=is_dev,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": new_location.path(),
                },
            },
        )
        return None


def package_url_from_pecl(pkg_name: str, version: str) -> str:
    return PackageURL(  # type: ignore
        type="pecl",
        namespace="",
        name=pkg_name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def _get_author(composer_package: PackagistPackageInfo) -> str | None:
    if not composer_package.get("authors"):
        return None

    authors: list[str] = []
    authors_dict = composer_package["authors"]
    for author_item in authors_dict:
        author: str = author_item["name"]
        if "email" in author_item:
            author_email = author_item["email"]
            author += f" <{author_email}>"
        authors.append(author)

    return ", ".join(authors)


def _set_health_metadata(
    package: Package,
    composer_package: PackagistPackageInfo,
    current_package: PackagistPackageInfo | None,
) -> None:
    package.health_metadata = HealthMetadata(
        latest_version=composer_package["version"],
        latest_version_created_at=composer_package["time"],
        artifact=_get_artifact_metadata(current_package) if current_package else None,
        authors=_get_author(composer_package),
    )


def _get_artifact_metadata(
    current_package: PackagistPackageInfo | None,
) -> Artifact | None:
    if current_package:
        dist_info = current_package.get("dist")
        if isinstance(dist_info, dict) and isinstance(dist_info.get("url"), str):
            digest_value = dist_info.get("shasum") or None
            return Artifact(
                url=dist_info["url"],
                integrity=Digest(
                    algorithm=infer_algorithm(digest_value),
                    value=digest_value,
                ),
            )
    return None


def _update_advisories(package: Package) -> None:
    pkg_advisories = advisories.get_vulnerabilities("composer", package.name, package.version)
    if pkg_advisories:
        package.advisories = pkg_advisories


def complete_package(package: Package) -> Package:
    _update_advisories(package)
    current_package = get_composer_package(package.name, package.version)
    composer_package = get_composer_package(package.name)

    if not composer_package:
        return package

    _set_health_metadata(package, composer_package, current_package)

    package.licenses = validate_licenses(composer_package["license"])

    return package



import logging
import re
from copy import (
    deepcopy,
)

from bs4 import (
    BeautifulSoup,
)
from bs4.element import (
    Tag,
)
from packageurl import (
    PackageURL,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)

LOGGER = logging.getLogger(__name__)


def parse_dotnet_config_executable(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    root = BeautifulSoup(reader.read_closer.read(), features="html.parser")
    net_dep = re.compile(r".NETFramework,Version=v(?P<version>[^\s,]*)")
    net_runtime = root.find("supportedruntime")
    if (
        isinstance(net_runtime, Tag)
        and isinstance(net_runtime.parent, Tag)
        and net_runtime.parent.name == "startup"
        and (runtime_info := net_runtime.get("sku", ""))
        and (version_match := net_dep.match(str(runtime_info)))
    ):
        version = version_match.group("version")

        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = net_runtime.sourceline
            location.dependency_type = DependencyType.DIRECT

        packages.append(
            Package(
                name="netframework",
                version=version,
                locations=[location],
                licenses=[],
                type=PackageType.DotnetPkg,
                language=Language.DOTNET,
                metadata=None,
                p_url=PackageURL(
                    type="nuget",
                    namespace="",
                    name="netframework",
                    version=version,
                    qualifiers={},
                    subpath="",
                ).to_string(),
            ),
        )

    return packages, []



import logging

from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.internal.collection import (
    toml,
)
from labels.internal.collection.types import (
    IndexedDict,
    IndexedList,
    ParsedValue,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.python.model import (
    PythonRequirementsEntry,
)
from labels.pkg.cataloger.python.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def parse_poetry_lock(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    _content = reader.read_closer.read()

    toml_content: IndexedDict[str, ParsedValue] = toml.parse_toml_with_tree_sitter(_content)

    packages = _parse_packages(toml_content, reader)
    relationships = _parse_relationships(toml_content, packages)

    return packages, relationships


def _parse_packages(
    toml_content: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
) -> list[Package]:
    packages = []
    toml_pkgs = toml_content.get("package")
    if not isinstance(toml_pkgs, IndexedList):
        return []
    for package in toml_pkgs:
        if not isinstance(package, IndexedDict):
            continue
        name: str = str(package.get("name", ""))
        version: str = str(package.get("version", ""))

        if not name or not version:
            continue

        p_url = package_url(name, version, package)  # type: ignore

        location = (
            _get_location(reader.location, package.position.start.line)
            if isinstance(package, IndexedDict)
            else reader.location
        )

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    found_by=None,
                    locations=[location],
                    language=Language.PYTHON,
                    p_url=p_url,
                    metadata=PythonRequirementsEntry(
                        name=name,
                        extras=[],
                        markers=p_url,
                    ),
                    licenses=[],
                    type=PackageType.PythonPkg,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages


def _get_dependencies(
    package: ParsedValue,
    packages: list[Package],
) -> tuple[Package | None, IndexedDict[str, ParsedValue]] | None:
    if not isinstance(package, IndexedDict):
        return None
    _pkg = next((pkg for pkg in packages if pkg.name == package["name"]), None)
    deps = package.get("dependencies")
    if not isinstance(deps, IndexedDict):
        return None
    return _pkg, deps


def _parse_relationships(
    toml_content: IndexedDict[str, ParsedValue],
    packages: list[Package],
) -> list[Relationship]:
    relationships = []
    _pkg: Package | None = None
    toml_pkgs = toml_content.get("package")
    if not isinstance(toml_pkgs, IndexedList):
        return []
    for package in toml_pkgs:
        results = _get_dependencies(package, packages)
        if results:
            _pkg, deps = results
            dependencies = list(deps.keys())

        if _pkg and dependencies:
            for dep in dependencies:
                dep_pkg = next((pkg for pkg in packages if pkg.name == dep), None)
                if dep_pkg:
                    relationships.append(
                        Relationship(
                            from_=dep_pkg,
                            to_=_pkg,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        ),
                    )

    return relationships



import logging
from collections.abc import Generator
from contextlib import (
    suppress,
)
from copy import (
    deepcopy,
)

import requirements
from pydantic import (
    ValidationError,
)
from requirements.requirement import (
    Requirement,
)
from univers.versions import (
    InvalidVersion,
    PypiVersion,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.file.scope import Scope
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.python.model import (
    PythonRequirementsEntry,
)
from labels.pkg.cataloger.python.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def get_dep_version_range(dep_specs: list[tuple[str, str]]) -> str:
    version_obj = PypiVersion(dep_specs[0][1])
    for _, version in dep_specs[1:]:
        version_obj = max(PypiVersion(version), version_obj)

    return version_obj.string


def get_parsed_dependency(line: str) -> tuple[str, str, Requirement, bool] | None:
    with suppress(Exception):
        parsed_dep = next(iter(requirements.parse(line)))
        if not parsed_dep.specs:
            return None
        try:
            version = get_dep_version_range(parsed_dep.specs)
        except InvalidVersion:
            return None
        is_dev: bool = False
        if parsed_dep.extras and any("dev" in extra.lower() for extra in parsed_dep.extras):
            is_dev = True
        return str(parsed_dep.name), version, parsed_dep, is_dev
    return None


def split_lines_requirements(
    content: str,
) -> Generator[tuple[int, str], None, None]:
    last_line = ""
    line_number = 1
    for index, raw_line in enumerate(content.splitlines(), 1):
        if not last_line:
            line_number = index
        line = trim_requirements_txt_line(raw_line)
        if last_line != "":
            line = last_line + line
            last_line = ""
        if line.endswith("\\"):
            last_line += line.rstrip("\\")
            continue
        if not line:
            continue

        if any(
            (
                line.startswith("-e"),
                line.startswith("-r"),
                line.startswith("--requirements"),
            ),
        ):
            continue

        yield line_number, line


def parse_requirements_txt(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []

    line_number = 1

    try:
        content = reader.read_closer.read()
    except UnicodeDecodeError:
        return packages, []

    for line_number, line in split_lines_requirements(content):
        parsed_dep = get_parsed_dependency(line)

        # Avoid parsing big txt files that have nothing to do with pip
        if not parsed_dep:
            continue

        product, version, req, is_dev = parsed_dep

        if not product or not version:
            continue

        p_url = package_url(product, version, None)
        current_location = deepcopy(reader.location)
        if current_location.coordinates:
            current_location.scope = Scope.DEV if is_dev else Scope.PROD
            current_location.coordinates.line = line_number

        try:
            packages.append(
                Package(
                    name=product,
                    version=version,
                    found_by=None,
                    locations=[current_location],
                    language=Language.PYTHON,
                    p_url=p_url,
                    metadata=PythonRequirementsEntry(
                        name=str(req.name),
                        extras=req.extras,
                        version_constraint=",".join(f"{s[0]} {s[1]}" for s in req.specs)
                        if req.specs
                        else "",
                        markers=p_url,
                    ),
                    licenses=[],
                    type=PackageType.PythonPkg,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": current_location.path(),
                    },
                },
            )
            continue

    return packages, []


def remove_trailing_comment(line: str) -> str:
    parts = line.split("#", 1)
    if len(parts) < 2:
        # there aren't any comments
        return line
    return parts[0]


def parse_url(line: str) -> str:
    parts = line.split("@")

    if len(parts) > 1:
        desired_index = -1

        for index, raw_part in enumerate(parts):
            part = "".join([char for char in raw_part if char.isalnum()])

            if part.startswith("git"):
                desired_index = index
                break

        if desired_index != -1:
            return "@".join(parts[desired_index:]).strip()

    return ""


def trim_requirements_txt_line(line: str) -> str:
    line = line.strip()

    return remove_trailing_comment(line)



import base64
import logging
from typing import (
    NamedTuple,
)

import aioboto3  # type: ignore
import botocore

from labels.sources.docker import (
    ImageMetadata,
    get_docker_image,
)

LOGGER = logging.getLogger(__name__)


class AwsCredentials(NamedTuple):
    access_key_id: str
    secret_access_key: str
    session_token: str | None


class AwsRole(NamedTuple):
    external_id: str | None
    role: str


async def get_credentials(credentials: AwsRole) -> AwsCredentials | None:
    try:
        async with aioboto3.Session().client(service_name="sts") as sts_client:
            response = await sts_client.assume_role(
                ExternalId=credentials.external_id,
                RoleArn=credentials.role,
                RoleSessionName="FluidAttacksRoleVerification",
            )
            aws_credentials = AwsCredentials(
                access_key_id=response["Credentials"]["AccessKeyId"],
                secret_access_key=response["Credentials"]["SecretAccessKey"],
                session_token=response["Credentials"]["SessionToken"],
            )
    except botocore.exceptions.ClientError:
        LOGGER.exception("We found problems on your AWS credentials")
        raise
    else:
        return aws_credentials


async def run_boto3_fun(  # noqa: PLR0913
    credentials: AwsCredentials,
    service: str,
    function: str,
    region: str | None = None,
    parameters: dict[str, object] | None = None,
    paginated_results_key: str | None = None,
) -> dict[str, dict | list]:
    try:
        session = aioboto3.Session(
            aws_access_key_id=credentials.access_key_id,
            aws_secret_access_key=credentials.secret_access_key,
            aws_session_token=credentials.session_token,
        )
        async with session.client(service, region) as client:
            if paginated_results_key and client.can_paginate(function):
                paginator = client.get_paginator(function)
                page_iterator = paginator.paginate(**(parameters or {}))
                merged_pages_list = []
                async for page in page_iterator:
                    merged_pages_list.extend(page[paginated_results_key])
                return {paginated_results_key: merged_pages_list}
            return await getattr(client, function)(**(parameters or {}))
    except botocore.exceptions.ClientError:
        LOGGER.exception("We found problems on your AWS credentials")
        raise


async def get_token(credentials: AwsCredentials) -> str:
    get_authorization_token = await run_boto3_fun(
        credentials=credentials,
        service="ecr",
        function="get_authorization_token",
    )

    auth_resp = get_authorization_token["authorizationData"][0]
    token = auth_resp["authorizationToken"]
    _username, password = base64.b64decode(token).decode("utf-8").split(":")
    return password


async def ecr_connection(
    role: AwsRole,
    image_uri: str,
) -> tuple[str, ImageMetadata | None]:
    credentials = await get_credentials(role)
    token = ""
    image_metadata = None
    if credentials:
        token = await get_token(credentials)
        image_ref = f"docker://{image_uri}"
        aws_creds = f"AWS:{token}"
        image_metadata = get_docker_image(
            image_ref,
            aws_creds=aws_creds,
        )

    return token, image_metadata



from fnmatch import (
    fnmatch,
)

import reactivex
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)

from labels.pkg.cataloger.generic.cataloger import (
    Request,
)
from labels.pkg.cataloger.rust.parse_cargo_lock import (
    parse_cargo_lock,
)
from labels.pkg.cataloger.rust.parse_cargo_toml import (
    parse_cargo_toml,
)


def on_next_rust(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if any(fnmatch(value, pattern) for pattern in ("**/Cargo.lock", "Cargo.lock")):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_cargo_lock,
                            parser_name="parse-rust-cargo-lock",
                        ),
                    )
                elif any(fnmatch(value, pattern) for pattern in ("**/Cargo.toml", "Cargo.toml")):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_cargo_toml,
                            parser_name="parse-rust-cargo-toml",
                        ),
                    )
            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



import logging
from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from labels.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from labels.internal.collection.types import (
    IndexedDict,
    ParsedValue,
)
from labels.model.core import (
    Package,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.php.model import (
    PhpComposerInstalledEntry,
    PhpComposerLockEntry,
)
from labels.pkg.cataloger.php.package import (
    new_package_from_composer,
)

LOGGER = logging.getLogger(__name__)
EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def parse_composer_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    packages: list[Package] = []
    relationships: list[Relationship] = []
    pkgs: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        package_json.get("packages", EMPTY_DICT),
    )
    pkgs_dev: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        package_json.get("packages-dev", EMPTY_DICT),
    )
    for is_dev, package in [
        *[(False, x) for x in pkgs],
        *[(True, x) for x in pkgs_dev],
    ]:
        if not isinstance(package, IndexedDict):
            continue
        new_location = deepcopy(reader.location)

        if pkg := new_package_from_composer(package, new_location, is_dev=is_dev):
            packages.append(pkg)

    for parsed_pkg in packages:
        if not isinstance(parsed_pkg.metadata, PhpComposerInstalledEntry | PhpComposerLockEntry):
            continue
        for dep_name in parsed_pkg.metadata.require or []:
            package_dep = next(
                (x for x in packages if x.name == dep_name),
                None,
            )
            if package_dep:
                relationships.append(
                    Relationship(
                        from_=parsed_pkg,
                        to_=package_dep,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    ),
                )
    return packages, relationships



from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.file.scope import (
    Scope,
)
from labels.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from labels.internal.collection.types import (
    IndexedDict,
    IndexedList,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.cpp.package import (
    package_url,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)


def format_conan_lock_dep(
    dep_info: str,
    location: Location,
    *,
    is_dev: bool = False,
) -> Package:
    product, version = dep_info.split("/")
    version = version.split("#")[0]
    location.scope = Scope.DEV if is_dev else Scope.PROD
    return Package(
        name=product,
        version=version,
        type=PackageType.ConanPkg,
        locations=[location],
        p_url=package_url(product, version),
        metadata=None,
        language=Language.CPP,
        licenses=[],
        is_dev=is_dev,
    )


def parse_conan_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    conan_file = cast(
        IndexedDict[str, IndexedList[str]],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    for index, dep_line in enumerate(conan_file.get("requires", [])):
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = (
                conan_file["requires"]
                .get_position(
                    index,
                )
                .start.line
            )
        packages.append(
            format_conan_lock_dep(dep_line, location),
        )
    for index, dep_line in enumerate(conan_file.get("build_requires", [])):
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = conan_file["build_requires"].get_position(index).start.line
        packages.append(
            format_conan_lock_dep(dep_line, location, is_dev=True),
        )
    return packages, []



from typing import cast

from spdx_matcher import analyse_license_text

LICENSES_FILE_NAMES = [
    "AL2.0",
    "COPYING",
    "COPYING.md",
    "COPYING.markdown",
    "COPYING.txt",
    "LGPL2.1",
    "LICENCE",
    "LICENCE.md",
    "LICENCE.markdown",
    "licence.txt",
    "LICENCE.txt",
    "LICENSE",
    "LICENSE.md",
    "LICENSE.markdown",
    "LICENSE.txt",
    "LICENSE-2.0.txt",
    "LICENCE-2.0.txt",
    "LICENSE-APACHE",
    "LICENCE-APACHE",
    "LICENSE-APACHE-2.0.txt",
    "LICENCE-APACHE-2.0.txt",
    "LICENSE-MIT",
    "LICENCE-MIT",
    "LICENSE.MIT",
    "LICENCE.MIT",
    "LICENSE.code",
    "LICENCE.code",
    "LICENSE.docs",
    "LICENCE.docs",
    "LICENSE.rst",
    "LICENCE.rst",
    "MIT-LICENSE",
    "MIT-LICENCE",
    "MIT-LICENSE.md",
    "MIT-LICENCE.md",
    "MIT-LICENSE.markdown",
    "MIT-LICENCE.markdown",
    "MIT-LICENSE.txt",
    "MIT-LICENCE.txt",
    "MIT_LICENSE",
    "MIT_LICENCE",
    "UNLICENSE",
    "UNLICENCE",
]


def parse_license(reader: str) -> list[str]:
    licenses_detected, _ = cast(
        tuple[dict[str, dict[str, str]], float],
        analyse_license_text(reader),
    )
    return list(licenses_detected["licenses"].keys())



from pydantic import (
    BaseModel,
)


class DigestAlgorithm(BaseModel):
    algorithm: int

    def __str__(self) -> str:
        return self.get_algorithm_name()

    def get_algorithm_name(self) -> str:
        return {
            1: "md5",
            2: "sha1",
            3: "ripemd160",
            5: "md2",
            6: "tiger192",
            7: "haval-5-160",
            8: "sha256",
            9: "sha384",
            10: "sha512",
            11: "sha224",
        }.get(self.algorithm, "unknown-digest-algorithm")



import logging
from typing import (
    Any,
)

import phpserialize
from pydantic import (
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.php.package import (
    package_url_from_pecl,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def php_to_python(obj: Any) -> Any:
    if isinstance(obj, phpserialize.phpobject):
        return {k: php_to_python(v) for k, v in obj.__dict__.items()}
    if isinstance(obj, dict):
        return {php_to_python(k): php_to_python(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [php_to_python(i) for i in obj]

    return obj


def parse_pecl_serialized(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []

    unserialized_data = phpserialize.loads(reader.read_closer.read().encode(), decode_strings=True)
    parsed_data = php_to_python(unserialized_data)
    name = parsed_data.get("name")
    version = parsed_data.get("version", {}).get("release")

    if not name or not version:
        return [], []

    try:
        packages.append(
            Package(
                name=name,
                version=version,
                locations=[reader.location],
                language=Language.PHP,
                licenses=[],
                type=PackageType.PhpPeclPkg,
                metadata=None,
                p_url=package_url_from_pecl(name, version),
            ),
        )
        return packages, relationships
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": reader.location.path(),
                },
            },
        )
        return [], []



import logging
import re
from copy import (
    deepcopy,
)

from pydantic import (
    BaseModel,
    ValidationError,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)
from labels.pkg.cataloger.java.model import (
    JavaArchive,
    JavaPomProject,
)
from labels.pkg.cataloger.java.package import (
    package_url,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)
# Constants
QUOTE = r'["\']'
NL = r"(\n?\s*)?"
TEXT = r'[^"\']+'
RE_LINE_COMMENT: re.Pattern[str] = re.compile(r"^.*" rf"{NL}//" r".*$")  # noqa: ISC001


def is_comment(line: str) -> bool:
    return (
        line.strip().startswith("//")
        or line.strip().startswith("/*")
        or line.strip().endswith("*/")
    )


class LockFileDependency(BaseModel):
    group: str
    name: str
    version: str
    line: int | None = None


def create_java_archive(
    dependency: LockFileDependency,
    name: str,
    version: str,
) -> JavaArchive:
    return JavaArchive(
        pom_project=JavaPomProject(
            group_id=dependency.group,
            name=name,
            artifact_id=name,
            version=version,
        ),
    )


def create_package(
    name: str,
    version: str,
    location: Location,
    archive: JavaArchive,
) -> Package:
    return Package(
        name=name,
        version=version,
        locations=[location],
        language=Language.JAVA,
        type=PackageType.JavaPkg,
        metadata=archive,
        p_url=package_url(name, version, archive),
        licenses=[],
    )


def avoid_cmt(line: str, *, is_block_cmt: bool) -> tuple[str, bool]:
    if RE_LINE_COMMENT.match(line):
        line = line.split("//", 1)[0]
    if is_block_cmt:
        if "*/" in line:
            is_block_cmt = False
            line = line.split("*/", 1).pop()
        else:
            return "", is_block_cmt
    if "/*" in line:
        line_cmt_open = line.split("/*", 1)[0]
        if "*/" in line:
            line = line_cmt_open + line.split("*/", 1).pop()
        else:
            line = line_cmt_open
            is_block_cmt = True
    return line, is_block_cmt


def get_line_number(content: str, match_start: int) -> int:
    return content[:match_start].count("\n") + 2


def extract_gradle_configs(content: str) -> set[str]:
    config_pattern = re.compile(r"configurations\s*\{([^}]+)\}", re.DOTALL)
    custom_config_pattern = re.compile(r"\s*(\w+)\s*")
    configs = {
        "runtimeOnly",
        "api",
        "compile",
        "compileOnly",
        "implementation",
    }

    config_blocks = config_pattern.findall(content)
    for block in config_blocks:
        custom_configs = custom_config_pattern.findall(block)
        configs.update(custom_configs)

    return configs


def build_regex_with_configs(configs: set[str]) -> dict[str, re.Pattern[str]]:
    config_pattern = "|".join(configs)

    return {
        "RE_GRADLE_A": re.compile(
            rf"^{NL}(?:{config_pattern}){NL}[(]?{NL}"
            rf"group{NL}:{NL}{QUOTE}(?P<group>{TEXT}){QUOTE}{NL},"
            rf"{NL}name{NL}:{NL}{QUOTE}(?P<name>{TEXT}){QUOTE}{NL}"
            rf"(?:,{NL}version{NL}:{NL}{QUOTE}(?P<version>{TEXT}){QUOTE}{NL})"
            rf"?.*$",
        ),
        "RE_GRADLE_B": re.compile(
            rf"^.*{NL}(?:{config_pattern}){NL}[(]?{NL}{QUOTE}(?P<statement>{TEXT}){QUOTE}",
        ),
        "RE_GRADLE_C": re.compile(
            rf"{NL}(?:{config_pattern}){NL}\("
            rf"{NL}{QUOTE}(?P<statement>{TEXT}){QUOTE}{NL}\)"
            rf"{NL}{{({NL})version{NL}{{({NL})strictly{NL}\({NL}"
            rf"{QUOTE}(?P<version>{TEXT}){QUOTE}{NL}\){NL}}}{NL}}}",
            re.DOTALL,
        ),
        "BLOCK": re.compile(
            rf"{NL}(?:{config_pattern}){NL}\("
            rf"{NL}{QUOTE}(?P<statement>{TEXT}){QUOTE}{NL}\)"
            rf"{NL}\{{(.*?version{NL}\{{.*?\}}){NL}\}}",
            re.DOTALL,
        ),
        "VERSION": re.compile(
            rf"version{NL}{{({NL})strictly{NL}\("
            rf"{NL}{QUOTE}(?P<version>{TEXT}){QUOTE}{NL}\){NL}}}",
            re.DOTALL,
        ),
    }


def get_block_deps(content: str, regexes: dict[str, re.Pattern[str]]) -> list[LockFileDependency]:
    dependencies = []

    for block in regexes["BLOCK"].finditer(content):
        product = block.group("statement")
        hit = regexes["VERSION"].search(block.group())
        if not hit:
            continue

        version = hit.group("version")
        if version == "":
            continue

        line_no = get_line_number(content, block.start())

        dependencies.append(
            LockFileDependency(
                group=product.split(":")[0],
                name=product,
                version=version,
                line=line_no,
            ),
        )
    return dependencies


def parse_dependencies(reader: LocationReadCloser) -> list[LockFileDependency]:
    content = reader.read_closer.read()
    configs = extract_gradle_configs(content)
    maven_regexes = build_regex_with_configs(configs)
    dependencies: list[LockFileDependency] = get_block_deps(content, maven_regexes)

    is_block_cmt = False
    for line_no, raw_line in enumerate(content.splitlines(), start=1):
        line, is_block_cmt = avoid_cmt(raw_line, is_block_cmt=is_block_cmt)

        if match := maven_regexes["RE_GRADLE_A"].match(line):
            group = match.group("group")
            product = group + ":" + match.group("name")
            version = match.group("version") or ""
        elif match := maven_regexes["RE_GRADLE_B"].match(line):
            statement = match.group("statement")
            product, version = (
                statement.rsplit(":", maxsplit=1) if statement.count(":") >= 2 else (statement, "")
            )
            group = product.split(":")[0]
        else:
            continue

        # Assuming a wildcard in Maven if the version is not found can
        # result in issues.
        # https://gitlab.com/fluidattacks/universe/-/issues/5635
        if version == "" or re.match(r"\${.*}", version):
            continue

        dependencies.append(
            LockFileDependency(
                group=group,
                name=product,
                version=version,
                line=line_no,
            ),
        )

    return dependencies


def create_packages(
    dependencies: list[LockFileDependency],
    reader_location: Location,
) -> list[Package]:
    packages: list[Package] = []

    for dependency in dependencies:
        name = dependency.name
        version = dependency.version

        if not name or not version:
            continue

        location = deepcopy(reader_location)
        if location.coordinates:
            location.coordinates.line = dependency.line

        archive = create_java_archive(dependency, name, version)

        try:
            package = create_package(name, version, location, archive)
            packages.append(package)
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )

    return packages


def parse_gradle(
    _resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    dependencies = parse_dependencies(reader)
    packages = create_packages(dependencies, reader.location)
    return packages, []



import atexit
import logging
import os
import sqlite3
import subprocess  # nosec
import sys
from pathlib import Path
from typing import (
    Literal,
    cast,
)

import boto3
from botocore import (
    UNSIGNED,
)
from botocore.config import (
    Config,
)
from botocore.exceptions import (
    ClientError,
)
from platformdirs import (
    user_data_dir,
)
from tqdm import (
    tqdm,
)

LOGGER = logging.getLogger(__name__)

BUCKET_NAME = "fluidattacks.public.storage"
DB_NAME = "vulnerability.db"
FILE_KEY = f"sbom/{DB_NAME}.zst"
CONFIG_DIRECTORY = user_data_dir(
    appname="fluid-labels",
    appauthor="fluidattacks",
    ensure_exists=True,
)
DB_PATH = os.path.join(CONFIG_DIRECTORY, DB_NAME)
DB_COMPRESSED_PATH = f"{DB_PATH}.zst"

S3_SERVICE_NAME: Literal["s3"] = "s3"
S3_CLIENT = boto3.client(
    service_name=S3_SERVICE_NAME,
    config=Config(
        region_name="us-east-1",
        signature_version=UNSIGNED,  # type: ignore[misc]
    ),
)


def _download_database_file(download_size: float) -> None:
    LOGGER.info("⬇️ Downloading advisories database")
    with tqdm(
        leave=False,
        total=download_size,
        unit="B",
        unit_scale=True,
        file=sys.stdout,
    ) as progress_bar:
        S3_CLIENT.download_file(
            Bucket=BUCKET_NAME,
            Callback=progress_bar.update,
            Filename=DB_COMPRESSED_PATH,
            Key=FILE_KEY,
        )


def _decompress_database_file() -> None:
    LOGGER.info("🗜️ Decompressing advisories database")
    with subprocess.Popen(  # noqa: S603
        ["zstd", "-d", "-f", DB_COMPRESSED_PATH],  # noqa:S607
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as process:
        _, stderr = process.communicate()
        if cast(int, process.returncode) != 0:
            raise RuntimeError(stderr.decode())


def _is_database_available() -> bool:
    local_database_exists = Path(DB_PATH).is_file()

    try:
        db_metadata = S3_CLIENT.head_object(Bucket=BUCKET_NAME, Key=FILE_KEY)
        up_to_date = (
            local_database_exists
            and Path(DB_PATH).stat().st_mtime >= db_metadata["LastModified"].timestamp()
        )

        if up_to_date:
            LOGGER.info("✅ Advisories database is up to date")
            return True

        _download_database_file(db_metadata["ContentLength"])
        _decompress_database_file()
        Path(DB_COMPRESSED_PATH).unlink()
        return True
    except (ClientError, RuntimeError):  # type: ignore[misc]
        if local_database_exists:
            LOGGER.warning(
                "⚠️ Advisories may be outdated, unable to update database",
            )
            return True

        LOGGER.exception(
            "❌ Advisories won't be included, unable to download database",
        )
        return False


class Database:
    def __init__(self) -> None:
        self.connection: sqlite3.Connection | None = None

    def initialize(self) -> None:
        if self.connection is None and _is_database_available():
            self.connection = sqlite3.connect(
                DB_PATH,
                # Should be OK as we are only reading, not writing
                check_same_thread=False,
            )
            atexit.register(self.connection.close)

    def get_connection(self) -> sqlite3.Connection | None:
        return self.connection


DATABASE = Database()



import re
from copy import (
    deepcopy,
)
from typing import (
    cast,
)

from labels.artifact.relationship import (
    Relationship,
)
from labels.file.dependency_type import (
    DependencyType,
)
from labels.file.location import (
    Location,
)
from labels.file.location_read_closer import (
    LocationReadCloser,
)
from labels.file.resolver import (
    Resolver,
)
from labels.file.scope import (
    Scope,
)
from labels.model.core import (
    Language,
    Package,
    PackageType,
)
from labels.pkg.cataloger.cpp.package import (
    package_url,
)
from labels.pkg.cataloger.generic.parser import (
    Environment,
)


def get_conan_dep_info(dep_line: str) -> tuple[str, str]:
    product, version = re.sub(r"[\"\]\[]", "", dep_line).strip().split("@")[0].split("/")
    if "," in version:
        version = re.sub(r",(?=[<>=])", " ", version).split(",")[0]
    return product, version


def format_conan_lock_dep(
    dep_info: str,
    location: Location,
    *,
    is_dev: bool = False,
) -> Package | None:
    regex = r"^([a-zA-Z0-9\-_]+)\/([^\s@]+)"
    match = re.match(regex, dep_info)
    location.scope = Scope.DEV if is_dev else Scope.PROD
    if match:
        name, version = cast(tuple[str, str], match.groups())
        return Package(
            name=name,
            version=version,
            type=PackageType.ConanPkg,
            locations=[location],
            p_url=package_url(name, version),
            metadata=None,
            language=Language.CPP,
            licenses=[],
            is_dev=is_dev,
        )
    return None


def build_location(
    reader: LocationReadCloser,
    line_number: int,
    *,
    is_dev: bool,
) -> Location:
    location = deepcopy(reader.location)
    location.scope = Scope.DEV if is_dev else Scope.PROD
    if location.coordinates:
        location.coordinates.line = line_number
        location.dependency_type = DependencyType.DIRECT
    return location


def parse_conan_file(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    line_deps: bool = False
    is_dev = False
    for line_number, line in enumerate(
        reader.read_closer.read().splitlines(),
        1,
    ):
        if re.search(r"^\[(tool|build)_requires\]$", line):
            line_deps = True
            is_dev = True
        elif not is_dev and line.startswith("[requires]"):
            line_deps = True
            is_dev = False
        elif line_deps:
            if not line or line.startswith("["):
                break
            pkg_name, pkg_version = get_conan_dep_info(line)
            location = build_location(reader, line_number, is_dev=is_dev)
            packages.append(
                Package(
                    name=pkg_name,
                    version=pkg_version,
                    type=PackageType.ConanPkg,
                    locations=[location],
                    p_url=package_url(pkg_name, pkg_version),
                    metadata=None,
                    language=Language.CPP,
                    licenses=[],
                    is_dev=is_dev,
                ),
            )
    return packages, []



import logging
from typing import TYPE_CHECKING

from labels.utils.strings import format_exception

if TYPE_CHECKING:
    from collections.abc import ItemsView
from typing import cast

from packageurl import PackageURL
from pydantic import ValidationError

from labels.artifact.relationship import Relationship
from labels.file.dependency_type import DependencyType
from labels.file.location import Location
from labels.file.location_read_closer import LocationReadCloser
from labels.file.resolver import Resolver
from labels.internal.collection.types import IndexedDict
from labels.internal.collection.yaml import parse_yaml_with_tree_sitter
from labels.model.core import Language, Package, PackageType
from labels.pkg.cataloger.generic.parser import Environment

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        location.dependency_type = DependencyType.DIRECT
        return location.model_copy(update=l_upd)
    return location


def _get_packages(
    reader: LocationReadCloser,
    dependencies: IndexedDict[str, str] | None,
) -> list[Package]:
    if dependencies is None:
        return []

    packages = []

    general_location = _get_location(reader.location, dependencies.position.start.line)
    items: ItemsView[str, IndexedDict[str, str] | str] = dependencies.items()
    for name, version in items:
        if not name or not isinstance(version, str) or not version:
            continue

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[general_location],
                    language=Language.DART,
                    licenses=[],
                    type=PackageType.DartPubPkg,
                    p_url=PackageURL(  # type: ignore
                        type="pub",
                        name=name,
                        version=version,
                    ).to_string(),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": general_location.path(),
                    },
                },
            )
            continue
    return packages


def parse_pubspec_yaml(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = cast(
        IndexedDict[str, IndexedDict[str, str]],
        parse_yaml_with_tree_sitter(reader.read_closer.read()),
    )
    deps: IndexedDict[str, str] | None = content.get("dependencies")
    dev_deps: IndexedDict[str, str] | None = content.get("dev_dependencies")
    packages = [
        *_get_packages(reader, deps),
        *_get_packages(reader, dev_deps),
    ]
    return packages, []
